********************************************************************
* Program....: DataClass.prg
* Compiler...: Visual FoxPro 07.00.0000.8853 for Windows
* Abstract...: Full set of VFP and SQL Data Classes
* ...........: 03/26: Modified to remove CRLF suppression in Memo fields
********************************************************************

********************************************************************
*** xCur -- Root Cursor Class **************************************
********************************************************************
DEFINE CLASS xCur AS CURSOR
  *** xCur::Property Listing *************************************
  cBehaviour      = ''       && Type of behaviour cursor will use
  cConnName       = ''       && Connection name cursor will use
  ****************************************************************
  cSelFields      = '*'      && Fields to be returned from the SELECT
  cTables         = ''       && Tables participating in the SELECT
  cPermWhere      = ''       && Filter clause to be appended in every ad hoc query
  cPermJoins      = ''       && Join clause to be appended in every ad hoc query
  cPermHaving     = ''       && Having clause to be appended in every ad hoc query
  cDefOrder       = ''       && Default Order by clause
  cDefGroupBy     = ''       && Default Group by clause
  ****************************************************************
  lUpdatable      = .T.      && Is cursor to be made updatable
  cUpdTable       = ''       && Table which is to be updated ( one table ONLY )
  cKeyField       = ''       && Name of key field ( Used for Update )
  cUpdFields      = ''       && List of Updatable Fields ( Default is ALL )
  ****************************************************************
  cLastWhere      = '0 = 1'  && Last WHERE clause used -- used for REQUERY
  lPopulate       = .F.      && Is the cursor to be populated with ALL data when opened
  lUseTblBuff     = .T.      && Should cursor use Table Buffering ( Default is yes )

  ****************************************************************
  *** xCur::Init *************************************************
  ****************************************************************
  FUNCTION INIT()
    WITH THIS
      *** If no alias specified -- assume same as class name
      .ALIAS     =  IIF( !EMPTY( .ALIAS ), .ALIAS, .NAME )
      *** If no source table specified -- assume same as Alias
      .cTables   =  IIF( !EMPTY( .cTables ), .cTables, .ALIAS )
      *** If no update table specified -- assume same as Source Table
      .cUpdTable = IIF( !EMPTY( .cUpdTable ), .cUpdTable, .cTables )
    ENDWITH
  ENDFUNC

  ****************************************************************
  *** xCur::Destroy **********************************************
  ****************************************************************
  FUNCTION DESTROY()
    WITH THIS
      *** Ensure no changes if cursor is to be closed
      IF USED( .ALIAS )
        IF CURSORGETPROP( 'Buffering', .ALIAS ) > 1
          TABLEREVERT( .T., .ALIAS )
        ENDIF
        USE IN ( .ALIAS )
      ENDIF
    ENDWITH
  ENDFUNC
ENDDEFINE

********************************************************************
*** Xcurvfp -- Vfp Cursor Class ************************************
********************************************************************
DEFINE CLASS xCurVFP AS xCur
  cBehaviour = 'BhvVFP'
  cConnName  = 'ConVFP'
ENDDEFINE

********************************************************************
*** Xcursql -- Sql Server Cursor Class *****************************
********************************************************************
DEFINE CLASS xCurSQL AS xCur
  cBehaviour = 'BhvSql'
  cConnName  = 'ConSql'
ENDDEFINE

*/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/*
*/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/*

********************************************************************
*** XDATMGR - Root Data Manager Class ******************************
********************************************************************
DEFINE CLASS xDatMgr AS xObjBase OLEPUBLIC

  DIMENSION aCons[1]    && Connection Collection
  nConCnt = 0

  ****************************************************************
  *** [E] ADDCONNECTION(): Attempt to add a connection object
  ****************************************************************
  FUNCTION AddConnection( tuConName )
    LOCAL loConDets, lcConn, lcClass, lcUid, lcPwd, lcDSN, loConObj, llIsConObj, llConnectOk
    *** Attempt to connect to the database, pass the connection details object
    *** Any errors will be logged in the lower level methods
    *** If nothing specified, just instantiate the Data Manager itself
    WITH THIS
      *** Check the input parameter type. If it's an object we assume that it is actually a 
      *** Connection definition object in the correct format, otherwise it is a connection name
      IF VARTYPE(tuConName) <> "O"
        *** We didn't get a connection object
        IF VARTYPE( tuConName ) = "C" AND ! EMPTY( tuConName )
          loConDets = .GetConnection( tuConName  )
          IF NOT loConDets.lStatus
            *** Could not find this connection so just return
            *** Error has already been logged
            RETURN .MakeResobj( .F. )
          ENDIF
        ELSE
          *** Invalid or no Connection Name passed
          .LOGERROR( 9014, tuConName, LOWER( PROGRAM()))
          RETURN .MakeResobj( .F. )
        ENDIF
      ELSE
        *** Just assign the passed in object to the local reference
        loConDets = tuConName
        *** But we also need to disconnect an existing connection 
        *** of the appropriate type - get it from the object
        loConObj = "o" + ALLTRIM( loConDets.cCon )
        *** And call the Data Manager's DisConnect method
        This.DisConnect( loConObj )
      ENDIF 

      *** Unpack the Connection details object
      WITH loConDets
        lcConn  = .cCon    && Connection Object Name (e.g. "CONSQL")
        lcBhv   = .cBhv    && Behavior Class to use (e.g. "BHVSQL")
        lcUid   = .cUID    && User ID (e.g. "sa" )
        lcPwd   = .cPWD    && Password (e.g. "change" )
        lcDSN   = .cDSN    && DSN to use (e.g. "SWPLUS" )
      ENDWITH

      *** Do we have one of these objects already?
      llIsConObj = .ChkConn( lcConn )
      IF ! llIsConObj
        *** Need to add the object
        .ADDOBJECT( 'o' + lcConn, 'x' + lcConn )
        *** But we need to check we have a connection object
        *** (AddObject ALWAYS returns .T.)
        llIsConObj = .ChkConn( lcConn )
        IF llIsConObj
          *** Add the new connection object to the connection
          .UpdCollection( 'ADD', lcConn )
        ENDIF
      ENDIF
      IF llIsConObj
        *** Get reference to the connection and try to connect
        loConObj = EVAL( 'This.o' + lcConn )
        llConnectOk = loConObj.CONNECT( lcUid, lcPwd, lcDSN )
        *** Check whether connection was successful
        IF llConnectOk
          *** Try and add the Behavior object too (saves time later)
          *** We really don't care if we succeed or not here
          *** This gets checked again when needed anyway
          loConObj.AddBehavior( lcBhv )
        ELSE
          *** Connect failed - Log Error & remove connection object
          .LOGERROR( 9017, lcDSN, LOWER( PROGRAM()))
          .REMOVEOBJECT( 'o' + lcConn )
          .UpdCollection( 'DEL', lcConn )
        ENDIF
      ELSE
        *** We couldn't add the connection object, let alone connect!
        THIS.LOGERROR( 9015, lcConn, LOWER( PROGRAM()))
        llConnectOk = .F.
      ENDIF
      *** Return success/failure
      RETURN THIS.MakeResobj( llConnectOk )
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [P] CHKCONN(): Check that connection manager is available
  ********************************************************************
  PROTECTED FUNCTION ChkConn(tcConn)
    LOCAL llOk
    *** Strip off leading 'o' if it was passed
    tcConn = STRTRAN( LOWER(tcConn), 'ocon', 'con' )
    *** Check connection manager is available - if not abort
    llOk = TYPE( 'This.o' + tcConn ) = 'O' AND ! ISNULL( EVAL( 'This.o' + tcConn ) )
    *** Return success/failure
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [E] CLOSETRANSACTIONS: Close open transactions on all connections
  ********************************************************************
  PROTECTED FUNCTION CloseTransactions( tnDS, tcAction )
    LOCAL lnCnt, loConObj, luRetVal, lnOldDS
    luRetVal = .T.
    *** Get into the right datasession
    lnOldDS = SET( "DataSession" )
    SET DATASESSION TO (tnDS)
    FOR lnCnt = 1 TO THIS.nConCnt
      *** Get the object Ref
      loConObj = EVALUATE( "This." + THIS.aCons[ lnCnt ] )
      *** And execute the appropriate action
      IF VARTYPE( loConObj ) = "O"
        luRetVal = luRetVal AND loConObj.&tcAction()
      ELSE
        THIS.LOGERROR( 9015, THIS.aCons[lnCnt], LOWER(PROGRAM()))
        luRetVal = NULL
      ENDIF
    NEXT
    *** Return the result here (force a NULL if an error occurred)
    IF NOT NVL( luRetVal, .F. )
      luRetVal = NULL
    ENDIF
    *** Restore DataSession
    SET DATASESSION TO (lnOldDS)
    RETURN luRetVal
  ENDFUNC
  
  ********************************************************************
  *** [E] DISCONNECT(): Execute the appropriate Disconnection method
  ********************************************************************
  FUNCTION DisConnect( tcConObjName )
  LOCAL loConObj
    WITH This
      *** Get the object reference to the connection manager
      loConObj = EVALUATE( "This." + tcConObjName )
      *** If a connection is open, close it
      IF loConObj.IsConnected()
        loConObj.Disconnect()
      ENDIF      
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [E] DO(): Data Manager Main Control Method
  ********************************************************************
  FUNCTION DO( tnDS, tcAction, toCursor, tuParm1, tuParm2, tuParm3, tuParm4, tuParm5 )
    #DEFINE NOPARMS 3  && Number of parameters that are not optional arguments
    LOCAL lnOldDS, lcConnObj, luRetVal, loConObj, lcBehaviour, loBhvObj

    *** Clean parameters
    tcAction = UPPER( ALLTRIM( tcAction ) )
    *** Extract connection name + behaviour from properties
    IF VARTYPE( toCursor ) = 'O'
      *** Use the cursor's properties to determine what's what
      loConObj  = 'o' + toCursor.cConnName
      lcBehaviour = toCursor.cBehaviour

      *** Check connection manager is available - if not abort
      IF ! THIS.ChkConn( loConObj )
        *** There is no way that we can determine what connection to use
        THIS.LOGERROR( 9015, toCursor.cConnName, LOWER( PROGRAM()))
        RETURN THIS.MakeResobj( NULL )
      ENDIF

      *** Check connection manager has desired behaviour + try to add if not
      IF TYPE( 'This.' + loConObj + '.o' + lcBehaviour )#'O' OR ISNULL( EVAL( 'This.' + loConObj + '.o' + lcBehaviour ) )
        *** It does not currently exist -- ask connection to add desired behaviour
        IF ! THIS.&loConObj..AddBehavior( lcBehaviour )
          *** Failed - so exit out right here!
          THIS.LOGERROR( 9016, toCursor.cBehaviour, LOWER( PROGRAM()))
          RETURN THIS.MakeResobj( NULL )
        ENDIF
      ENDIF
    ELSE
      *** Commit/RollBack can be called without a cursor
      *** Otherwise it's an Error!
      IF tcAction == "COMMIT" OR tcAction == "ROLLBACK"
        luRetVal = (THIS.CloseTransactions( tnDS, tcAction ))
      ELSE
        THIS.LOGERROR( 9005, tcCursor, LOWER( PROGRAM()))
        luRetVal = NULL
      ENDIF
      RETURN THIS.MakeResobj( luRetVal )
    ENDIF

    *** Save current DataSession + set to calling prog's
    lnOldDS = SET( 'DATASESSION' )
    SET DATASESSION TO ( tnDS )

    *** Check for special handling of action -- OTHERWISE handles all others
    DO CASE
      CASE tcAction == 'OPEN'
        *** Open a cursor
        luRetVal = THIS.OPEN( tnDS, toCursor )

      CASE tcAction == "EXECPROC"
        *** Execute a SQL 70 Stored Procedure
        luRetVal = THIS.&loConObj..&tcAction( tnDS, toCursor, tuParm1 )

      CASE tcAction == "GETLASTID"
        *** Retrieve the last Identity Value used on the connection
        *** Or for the specified VFP Table if one is passed
        luRetVal = THIS.&loConObj..&tcAction( tnDS, toCursor )

      OTHERWISE
        *** Call named method in behaviour object with passed parameters
        loBhvObj  = 'o' + lcBehaviour
        DO CASE
          CASE PCOUNT() = NOPARMS
            luRetVal = THIS.&loConObj..&loBhvObj..&tcAction.( toCursor )
          CASE PCOUNT() = NOPARMS + 1
            luRetVal = THIS.&loConObj..&loBhvObj..&tcAction.( toCursor, tuParm1 )
          CASE PCOUNT() = NOPARMS + 2
            luRetVal = THIS.&loConObj..&loBhvObj..&tcAction.( toCursor, tuParm1, tuParm2 )
          CASE PCOUNT() = NOPARMS + 3
            luRetVal = THIS.&loConObj..&loBhvObj..&tcAction.( toCursor, tuParm1, tuParm2, tuParm3 )
          CASE PCOUNT() = NOPARMS + 4
            luRetVal = THIS.&loConObj..&loBhvObj..&tcAction.( toCursor, tuParm1, tuParm2, tuParm3, tuParm4 )
          OTHERWISE
            luRetVal = THIS.&loConObj..&loBhvObj..&tcAction.( toCursor, tuParm1, tuParm2, tuParm3, tuParm4, tuParm5 )
        ENDCASE
    ENDCASE

    #UNDEF NOPARMS

    *** Restore DataSession
    SET DATASESSION TO ( lnOldDS )

    *** Return value to calling program
    RETURN THIS.MakeResobj( luRetVal )

  ENDFUNC

  ********************************************************************
  *** [P] GETCONNECTION(): Attempt to retrieve connection parameters from the table
  ********************************************************************
  PROTECTED FUNCTION GetConnection(tcConName)
    LOCAL llOk, loParm
    *** Try and open the connection lookup table
    IF NOT THIS.OpenLocalTable( 'connects' )
      THIS.LOGERROR( 9009, 'Connects.dbf', LOWER( PROGRAM()) )
      RETURN .F.
    ENDIF
    *** Create a Parameter Object for the return values
    loParm = CREATEOBJECT( "RELATION" )
    loParm.ADDPROPERTY( 'lStatus', .T. )

    *** All OK, so get the connection details
    IF SEEK( UPPER( ALLTRIM( tcConName )), 'connects', 'cconname')
      *** Read the connection details
      loParm.ADDPROPERTY( 'cDSN', UPPER(ALLTRIM( CONNECTS.ccondsce )))
      loParm.ADDPROPERTY( 'cUID', UPPER( ALLTRIM( CONNECTS.cconuid )))
      loParm.ADDPROPERTY( 'cPWD',UPPER( ALLTRIM( CONNECTS.cconpwd )))
      loParm.ADDPROPERTY( 'cCon','CON' + UPPER( ALLTRIM( CONNECTS.ccontype )))
      loParm.ADDPROPERTY( 'cBhv','BHV' + UPPER( ALLTRIM( CONNECTS.ccontype )))
    ELSE
      *** Connection name not recognised, log an error and return
      THIS.LOGERROR( 9014, tcConName, LOWER( PROGRAM()) )
      loParm.lStatus = .F.
    ENDIF
    RETURN loParm
  ENDFUNC

  ********************************************************************
  *** [E] GETHANDLE(): Return the connection handle for a specified connection type
  ********************************************************************
  FUNCTION GetHandle(tcType)
    LOCAL lnRetVal, lnCnt
    lnRetVal = 0
    FOR lnCnt = 1 TO THIS.nConCnt
      IF NOT LOWER(tcType) $ LOWER( THIS.aCons[lnCnt] )
        *** This is not the right type of connection
        LOOP
      ENDIF
      *** If we get here, we have a connection of the required type
      *** So get the connection handle
      loCon = EVALUATE( "This." + THIS.aCons[lnCnt])
      lnRetVal = loCon.GetHandle()
      EXIT
    NEXT
    RETURN lnRetVal
  ENDFUNC

  ********************************************************************
  *** [E] INIT(): Initializes the Data Manager
  ********************************************************************
  FUNCTION INIT( tcConName )
    LOCAL loConDets
    *** Call Parent Class Init to open the error message file
    DODEFAULT()
    IF ! EMPTY( tcConName )
      *** Try and Establish the connection, if one has been specified at startup
      THIS.AddConnection( tcConName )
    ENDIF
    *** Just Return out of here as normal
    RETURN
  ENDFUNC

  ********************************************************************
  *** [P] OPEN(): Opens a cursor - called when DataSets are initializing
  ********************************************************************
  PROTECTED FUNCTION OPEN( tnDS, toCursor, tlPopulate )
    LOCAL luRetVal, loConObj, loBhvObj, llPopulate
    *** Just pass - on to appropriate behaviour
    loConObj  = 'o' + toCursor.cConnName
    loBhvObj  = 'o' + toCursor.cBehaviour
    *** If not specified, check the cursor's lPopulate property
    llPopulate  = IIF( EMPTY( tlPopulate ), toCursor.lPopulate, .T. )
    *** Now call the Behavior Object Open method
    luRetVal    = THIS.&loConObj..&loBhvObj..OPEN( toCursor, llPopulate )
    *** Return value to calling program
    RETURN luRetVal
  ENDFUNC

  ********************************************************************
  *** [E] GETERRORS(): Extracts errors from the aErrors collection,
  *** packages them up in an object, and returns it
  ********************************************************************
  FUNCTION GetErrors()
    LOCAL loParamObj, lnCnt, loCon
    *** Get the errors for each connection object
    FOR lnCnt = 1 TO THIS.nConCnt
      *** Get the object Ref
      loCon = EVALUATE( "This." + THIS.aCons[ lnCnt ] )
      IF VARTYPE( loCon ) = "O"
        *** Add it's errors to this collection
        THIS.AddErrors( loCon )
      ENDIF
    NEXT
    *** And then do the Root Class Behavior
    loParamObj = DODEFAULT()
    RETURN loParamObj
  ENDFUNC

  ********************************************************************
  *** [P] MAKERESOBJ(): Creates and returns a result object, including any errors
  ********************************************************************
  PROTECTED FUNCTION MakeResobj(tuRetVal)
    WITH THIS
      *** Get the Error Object
      loResObj = THIS.GetErrors()
      *** And add the Return Value Property
      loResObj.ADDPROPERTY( 'uRetVal', tuRetVal )
      RETURN loResObj
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [P] UPDCOLLECTION(): Update the Connection collection
  ********************************************************************
  PROTECTED FUNCTION UpdCollection( tcAction, tcConName )
    LOCAL lcConn, lnConCnt
    WITH THIS
      lcConn = 'o' + LOWER( ALLTRIM( tcConName ) )
      IF tcAction = "ADD"
        *** How many connections currently?
        lnConCnt = ALEN( .aCons, 1 )
        *** If only one row, is it actually a connection?
        IF lnConCnt = 1 AND EMPTY( .aCons[ 1, 1 ] )
          *** Nope - set connection count = 1
          lnConCnt = 1
        ELSE
          *** We already have at least one connection
          lnConCnt = lnConCnt + 1
        ENDIF
        DIMENSION .aCons[ lnConCnt, 1 ]
        .aCons[ lnConCnt, 1 ] = lcConn
        .nConCnt = .nConCnt + 1
      ELSE
        *** Remove the connection. Find which row it is
        lnConCnt = ASCAN( .aCons, lcConn, 1, -1, -1, 9)
        *** And Delete it
        ADEL( .aCons, lnConCnt )
        *** Decrement the counter
        .nConCnt = IIF(.nConCnt > 1, .nConCnt - 1, 1 )
        *** And re-dimension the Array
        DIMENSION .aCons( .nConCnt )
      ENDIF
    ENDWITH
    RETURN
  ENDFUNC

  ********************************************************************
  *** [E] VALTOSQLSTR(): Format data for transmission to SQL Server
  ********************************************************************
  FUNCTION ValToSQLStr( tuExp, tcType )
    LOCAL lcRetVal, lcType
    lcType = IIF( VARTYPE( tcType ) = "C" AND ! EMPTY( tcType ), tcType, TYPE( 'tuExp' ))
    *** Convert from type to char
    DO CASE
      CASE ISNULL( tuExp )
        lcRetVal = "NULL"
      CASE lcType$'C'  && Character - get rid of Nulls and Embedded CRLF
        lcRetVal = CHRTRAN( tuExp, CHR(0) + CHR(13) + CHR(10), '')
        *** And embedded quotes (Convert Smart Quotes to standard first)
        lcRetVal = STRTRAN( lcRetVal, CHR(147), ['] )
        lcRetVal = STRTRAN( lcRetVal, CHR(148), ['] )
        *** Double quotes to pass them correctly
        lcRetVal = STRTRAN( lcRetVal, CHR(39), [''] )
        *** Remove trailing blanks and add enclosing quotes
        lcRetVal = "'" + RTRIM( lcRetVal ) + "'"
      CASE lcType$'I' OR ( lcType$'N' AND INT( tuExp ) = tuExp ) && Integer
        lcRetVal = ALLTRIM( PADL( INT( tuExp ), 16 ) )
      CASE lcType$'NY' && Numeric
        lcRetVal = CHRTRAN( ALLTRIM( PADL( tuExp, 32 ) ), ",", "." )
      CASE lcType$'L'  && Logical
        lcRetVal = IIF( !EMPTY( tuExp ), '1', '0' )
      CASE lcType$'DT' && Date mm/dd/yyyy or Datetime mm/dd/yyyy hh:mm:ss
        lcRetVal = "'" + PADL( MONTH( tuExp ), 2, '0' ) ;
          + '/' + PADL( DAY( tuExp ), 2, '0' ) ;
          + '/' + PADL( YEAR( tuExp ), 4, '0' )
        *** For DateTime just add on the time component
        IF lcType = 'T'
          lcRetVal = lcRetVal + ' ' + PADL( HOUR( tuExp ), 2, '0' ) ;
            + ':' + PADL( MINUTE( tuExp ), 2, '0' ) ;
            + ':' + PADL( SEC( tuExp ), 2, '0' )
        ENDIF
        *** Add the closing quote mark
        lcRetVal = lcRetVal + "'"
      CASE lcType$'M'  && Memo
        *** Remove embedded quotes & NULLS
        lcRetVal = CHRTRAN( tuExp, CHR(0), '')
        *** And embedded quotes (Convert Smart Quotes to standard first)
        lcRetVal = STRTRAN( lcRetVal, CHR(147), ['] )
        lcRetVal = STRTRAN( lcRetVal, CHR(148), ['] )
        *** Double quotes to pass them correctly
        lcRetVal = STRTRAN( lcRetVal, CHR(39), [''] )
        *** Remove trailing blanks and add enclosing quotes
        lcRetVal = "'" + RTRIM( lcRetVal ) + "'"
      OTHERWISE
        lcRetVal = ""
    ENDCASE
    *** Return value as character
    RETURN lcRetVal
  ENDFUNC

  ********************************************************************
  *** [E] VALTOORASTR(): Format data for transmission to Oracle
  ********************************************************************
  FUNCTION ValToORAStr(tuExp, tcType)
    LOCAL lcRetVal, lcType
    lcType = IIF( VARTYPE( tcType ) = "C" AND ! EMPTY( tcType ), tcType, TYPE( 'tuExp' ))
    *** Convert from type to char
    DO CASE
      CASE ISNULL( tuExp )
        lcRetVal = "NULL"
      CASE lcType$'C'  && Character
        *** Remove and CRLF characters
        lcRetVal = CHRTRAN( tuExp, CHR( 13 ) + CHR( 10 ), ' ' )
        *** Get rid of existing terminal quotes
        IF LEFT( lcRetVal, 1 ) $ ["']
          lcRetVal = SUBSTR( lcRetVal, 2)
        ENDIF
        IF RIGHT( lcRetVal, 1 ) $ ["']
          lcRetVal = LEFT( lcRetVal, LEN( lcRetVal )-1 )
        ENDIF
        *** Add enclosing Single Quotes
        lcRetVal = "'" + RTRIM( lcRetVal ) + "'"
      CASE lcType$'I' OR ( lcType$'N' AND INT( tuExp ) = tuExp ) && Integer
        lcRetVal = ALLTRIM( PADL( INT( tuExp ), 16 ) )
      CASE lcType$'NY' && Numeric
        lcRetVal = CHRTRAN( ALLTRIM( PADL( tuExp, 32 ) ), ",", "." )
      CASE lcType$'L'  && logical
        lcRetVal = IIF( !EMPTY( tuExp ), '1', '0' )
      CASE lcType$'D' && Date
        IF ! EMPTY( NVL( tuExp, {} ) )
          lcRetVal = "TO_DATE( '" + DTOS( tuExp ) + "', 'YYYYMMDD' )"
        ELSE
          lcRetVal = "NULL"
        ENDIF
      CASE lcType$'T' && Datetime
        IF ! EMPTY( NVL( tuExp, {} ) )
          lcRetVal = "TO_DATE( '" + TTOC( tuExp, 1 ) + "', 'YYYYMMDDHH24MISS' )"
        ELSE
          lcRetVal = "NULL"
        ENDIF
      CASE lcType$'M'  && Memo
        *** Remove embedded quotes & NULLS
        lcRetVal = CHRTRAN( tuExp, CHR(0), '')
        *** And embedded quotes (Convert Smart Quotes to standard first)
        lcRetVal = STRTRAN( lcRetVal, CHR(147), ['] )
        lcRetVal = STRTRAN( lcRetVal, CHR(148), ['] )
        *** Get rid of existing terminal quotes
        IF LEFT( lcRetVal, 1 ) $ ["']
          lcRetVal = SUBSTR( lcRetVal, 2)
        ENDIF
        IF RIGHT( lcRetVal, 1 ) $ ["']
          lcRetVal = LEFT( lcRetVal, LEN( lcRetVal )-1 )
        ENDIF
        *** Remove trailing blanks and add enclosing quotes
        lcRetVal = "'" + RTRIM( lcRetVal ) + "'"
      OTHERWISE
        lcRetVal = ""
    ENDCASE
    *** Return value as character
    RETURN lcRetVal
  ENDFUNC

  ********************************************************************
  *** [E] VALTOVFPSTR(): Format Data for transmission to VFP
  ********************************************************************
  FUNCTION ValToVFPStr( tuExp, tcType )
    LOCAL lcRetVal, lcType
    lcType = IIF( VARTYPE( tcType ) = "C" AND ! EMPTY( tcType ), tcType, TYPE( 'tuExp' ))
    *** Convert from type to char
    DO CASE
      CASE ISNULL( tuExp )
        lcRetVal = "NULL"
      CASE lcType$'C'  && Character
        *** Remove any NULL or CRLF characters
        lcRetVal = CHRTRAN( tuExp, CHR( 0 ) + CHR( 13 ) + CHR( 10 ), ' ' )
        lcRetVal = "[" + RTRIM( lcRetVal ) + "]"
      CASE lcType$'I' OR ( lcType$'N' AND INT( tuExp ) = tuExp ) && Integer
        lcRetVal = ALLTRIM( PADL( INT( tuExp ), 16 ) )
      CASE lcType$'NY' && Numeric
        lcRetVal = CHRTRAN( ALLTRIM( PADL( tuExp, 32 ) ), ",", "." )
      CASE lcType$'L'  && Logical
        lcRetVal = IIF( !EMPTY( tuExp ), '.T.', '.F.' )
      CASE lcType$'D' && Date
        lcRetVal = "CTOD( '" + DTOC( tuExp ) + "' )"
      CASE lcType$'T' && Datetime
        lcRetVal = "CTOT( '" + TTOC( tuExp ) + "' )"
      CASE lcType$'M'  && Memo
        *** Remove embedded NULLS
        lcRetVal = CHRTRAN( tuExp, CHR(0), '')
        *** Remove trailing blanks and add enclosing brackets
        lcRetVal = "[" + RTRIM( lcRetVal ) + "]"
      OTHERWISE
        lcRetVal = ""
    ENDCASE
    *** Return value as character
    RETURN lcRetVal
  ENDFUNC

  ********************************************************************
  *** [E] SHOWERRORS(): Retrieve and display Errors
  *** USE ONLY WHEN DEBUGGING - SHOULD NEVER BE CALLED AT RUN TIME
  ********************************************************************
  FUNCTION ShowErrors()
    LOCAL lnSelect, loErr, lcTxt, lcDispTxt
    *** Save Work Area
    lnSelect = SELECT()
    loErr = THIS.GetErrors()
    lcTxt = ""
    IF loErr.nErrorCount = 0
      *** No errors to display
      lcDispTxt = "No Errors were recorded"
      MESSAGEBOX( lcDispTxt, 64, "Error Report")
      RETURN
    ELSE
      FOR lnCnt = 1 TO loErr.nErrorCount
        lcTxt = lcTxt + IIF( NOT EMPTY( lcTxt ), CHR(13)+CHR(10), "" )
        *** Error Number
        lcTxt = lcTxt + ALLTRIM( TRANSFORM( loErr.aErrors[ lnCnt,1 ] )) + ": "
        lcTxt = lcTxt + ALLTRIM( loErr.aErrors[ lnCnt, 2 ] )
        IF ! EMPTY( loErr.aErrors[ lnCnt, 3 ] )
          lcTxt = lcTxt + " (" + ALLTRIM( loErr.aErrors[ lnCnt, 3 ] ) + ")"
        ENDIF
      NEXT
      lcDispTxt = lcTxt + CHR(13) + CHR(10) + CHR(13) + CHR(10)
      lcDispTxt = lcDispTxt + "Save these errors to FILE (ErrTxt.txt)?"
    ENDIF
    *** Display Error Messages
    IF MESSAGEBOX( lcDispTxt, 32 + 4 + 256, "Error Report") = 6 &&YES
      *** Save to file if required
      lcTxt = "Errors Recorded at: " + TTOC( DATETIME() ) + CHR(13) + CHR(10) + lcTxt
      STRTOFILE( lcTxt + CHR(13) + CHR(10), "errTxt.txt", .T. )
    ENDIF
    *** Tidy Up
    SELECT (lnSelect)
    RETURN
  ENDFUNC

ENDDEFINE

*/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/*
*/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/*

********************************************************************
*** XBHVMGR -- ROOT DATA BEHAVIOUR CLASS ***************************
********************************************************************
DEFINE CLASS xBhvMgr AS xObjBase

  ********************************************************************
  *** [P] xBhvMgr::CHKUPDTYPE(): Determine the type of SQL update required
  ********************************************************************
  PROTECTED FUNCTION ChkUpdType( toCursor )
    LOCAL lcOldDel, loSel, llTblBuf, lnLastRec, llOk
    *** Ensure deleted is off
    lcOldDel = SET( 'DELETED' )
    SET DELETED OFF
    *** Note work area + select cursor
    loSel = CREATEOBJECT( 'xSelAlias', toCursor.ALIAS )
    GO TOP
    *** Loop thru changed records + update
    llOk = .T.
    DO WHILE NOT EOF()
      luVal = toCursor.cKeyField
      *** Is this a changed record?
      lcFldState =  THIS.IsFieldChanged( toCursor.ALIAS, toCursor.cUpdFields )
      IF ( "2" $ lcFldState OR "4" $ lcFldState )
        *** Check the type of change and act on it appropriately
        DO CASE
          CASE ! DELETED() AND EMPTY( NVL( &luVal, 0) )  && inserted record
            llOk = THIS.SendInsert( toCursor )
          CASE DELETED() AND EMPTY( NVL( &luVal, 0) )    && deleted appended record - ignore
            llOk = .T.
          CASE ! DELETED()                               && update record
            llOk = THIS.SENDUPDATE( toCursor )
          CASE   DELETED()                               && delete record
            llOk = THIS.SendDelete( toCursor )
          OTHERWISE  && unknown condition
            llOk = .F.
        ENDCASE
      ELSE
        llOk = .T.
      ENDIF
      *** Exit now if error
      * If error message passed - back OR error condition OR row buffering
      IF ! llOk
        EXIT
      ENDIF
      SKIP
    ENDDO
    *** Restore deleted status
    IF lcOldDel = "ON"
      SET DELETED ON
    ENDIF
    *** Return success / failure
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::DELETE():  Delete a record
  ********************************************************************
  PROTECTED FUNCTION DELETE( tcTable, tcWhereClause )
    LOCAL lcSqlStr, llOk
    *** Build delete string
    lcSqlStr = 'DELETE FROM ' + tcTable + ' WHERE ' + tcWhereClause
    *** Execute Delete string and force Override of Connection Exec Method
    llOk = THIS.PARENT.EXEC( 0, lcSqlStr, '', .F., .T. )
    *** Report Status in Dev Mode Only
    ASSERT llOk MESSAGE "Unable to Delete in Table " + tcTable
    *** Return success/failure
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::FIELDCOUNT(): Return the number of fields allowed to be updated
  ********************************************************************
  PROTECTED FUNCTION FieldCount( tcFields )
    LOCAL lnCount, lnCnt
    *** Returns the number of updatable fields
    lnCount = 0
    IF ! EMPTY( tcFields )
      FOR lnCnt = 1 TO OCCURS( ",", tcFields ) + 1
        IF ! ISNULL( THIS.GetItem( tcFields, lnCnt ) )
          lnCount = lnCount + 1
        ENDIF
      NEXT
    ENDIF
    RETURN ( lnCount )
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::GETLOCATEEXP(): Construct expression for SQL WHERE clause
  ********************************************************************
  PROTECTED FUNCTION GetLocateExp( toCursor )
    THIS.NotImp( toCursor, PROGRAM() )
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::INSERT(): Insert a record into the back end table
  ********************************************************************
  PROTECTED FUNCTION INSERT( tcTable, tcFlds, tcVals )
    LOCAL lcSqlStr, llOk

    *** Build insert string
    lcSqlStr = 'INSERT INTO ' + tcTable

    *** If both parameters passed - build full string
    IF ! EMPTY( tcVals )
      lcSqlStr = lcSqlStr + ' ( ' + tcFlds + ' ) VALUES ( ' + tcVals + ' )'
    ELSE  && only vals must have been supplied ( in first string )
      lcSqlStr = lcSqlStr + ' VALUES ( ' + tcFlds + ' )'
    ENDIF

    *** Execute Insert string and force Override of Connection Exec Method
    llOk = THIS.PARENT.EXEC( 0, lcSqlStr, '', .F., .T. )

    *** Show status when in Dev Mode
    ASSERT llOk MESSAGE "Unable to Insert into Table " + tcTable

    *** Return success/failure
    RETURN llOk

  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::ISCHANGED(): Determines whether a record has changes
  ********************************************************************
  PROTECTED FUNCTION IsChanged( toCursor, tlCurrentRec )
    LOCAL llChanged, lcFldState, lnLastRec, lcAlias

    *** Retrieve alias property depending on settings.
    lcAlias = toCursor.ALIAS

    DO CASE
      CASE ! toCursor.lUpdatable
        *** If the cursor is not updateable - cannot be any changes
        llChanged = .F.
      CASE ( INLIST(CURSORGETPROP( "Buffering", lcAlias ), 2, 3) OR !EMPTY( tlCurrentRec ) ) AND !EOF( lcAlias )  && row
        lcFldState = THIS.IsFieldChanged( lcAlias, toCursor.cUpdFields )
        llChanged = ( "2" $ lcFldState OR "4" $ lcFldState )
      CASE INLIST( CURSORGETPROP( "Buffering", lcAlias ), 4, 5)  && table
        *** If we are Table Buffered
        *** Save record pointer.!
        lnRecno = IIF( !EOF( lcAlias ), RECNO( lcAlias ), 0 )
        *** Need to check all records for changes
        llChanged = .F.
        lnLastRec = 0
        *** Loop thru changed records to see if any of the updateable fields have changed
        DO WHILE ! llChanged
          lnLastRec = GETNEXTMODIFIED( lnLastRec, lcAlias )
          IF lnLastRec = 0
            EXIT
          ENDIF
          GO ( lnLastRec ) IN ( lcAlias )
          lcFldState = THIS.IsFieldChanged( lcAlias, toCursor.cUpdFields )
          llChanged = ( "2" $ lcFldState OR "4" $ lcFldState )
        ENDDO
        *** Restore record pointer
        IF lnRecno # 0
          GO ( lnRecno ) IN ( lcAlias )
        ENDIF
      OTHERWISE
        llChanged = .F.
    ENDCASE

    *** Return changed status
    RETURN llChanged
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::ISFIELDCHANGED(): Determine whether changes are in
  **  fields or deletion status
  ********************************************************************
  PROTECTED FUNCTION IsFieldChanged( tcAlias, tcUpdateFlds )
    LOCAL lnCnt, lcField, lnChangeVal
    *** If no update field list is specified, or the default is there
    *** just get the entire row and return it
    lcUpDFlds = ALLTRIM( STRTRAN( tcUpdateFlds, "*" ) )
    IF EMPTY( lcUpDFlds )
      RETURN ( NVL( GETFLDSTATE( - 1, tcAlias ), "0" ) )
    ENDIF

    *** To reach this point the cursor has to be an updatable
    *** view.  Scan through the list of updatable fields
    *** determining whether any have changed.
    lnChangeVal = 1
    FOR lnCnt = 1 TO OCCURS( ',', lcUpDFlds ) + 1
      lcField = THIS.GetItem( lcUpDFlds, lnCnt )
      *** If we don't have a change already, check the field
      IF MOD( lnChangeVal, 2 ) # 0
        lnChangeVal = NVL( GETFLDSTATE( lcField, tcAlias ), 0 )
      ELSE
        EXIT
      ENDIF
    NEXT
    *** If no other changes, check deletion status of record too
    IF MOD( lnChangeVal, 2 ) # 0
      lnChangeVal = NVL( GETFLDSTATE( 0, tcAlias ), 0 )
    ENDIF
    RETURN ( ALLTRIM( STR( lnChangeVal ) ) )
  ENDFUNC

  ********************************************************************
  *** [E] xBhvMgr::LOOKUPEXP(): Run query and return the specified field
  *** from the first record in the result set
  ********************************************************************
  FUNCTION LookupExp( toCursor, tcWhereExp, tcRetExp )
    LOCAL lnSelect, lcTmpCur, lcTable, lcStr, luSqlVal
    *** Make note of current work area
    lnSelect = SELECT()
    *** Perform select into temporary cursor
    lcTmpCur = SYS( 2015 )
    *** Get Table to Look Up into (use Joins if nothing in Tables)
    lcTable = IIF( EMPTY( toCursor.cTables ), toCursor.cPermJoins, toCursor.cTables )
    *** Build Sql String
    lcStr = ''
    lcStr = lcStr + 'SELECT ' + tcRetExp + ' SqlVal'
    lcStr = lcStr + ' FROM ' + lcTable
    IF !EMPTY( tcWhereExp )
      lcStr = lcStr + ' WHERE ' + tcWhereExp
    ENDIF
    THIS.PARENT.EXEC( SET( 'DATASESSION' ), lcStr, lcTmpCur, .T. )
    *** Extract value, close cursor + restore work area
    IF RECCOUNT( lcTmpCur ) > 0
      *** Extract value, close cursor + restore work area
      luSqlVal = EVAL( lcTmpCur + '.SqlVal' )
    ELSE
      luSqlVal = ""
    ENDIF
    USE IN ( lcTmpCur )
    SELECT ( lnSelect )
    *** Return value
    RETURN luSqlVal
  ENDFUNC

  ********************************************************************
  *** [E] xBhvMgr::OPEN(): Attempts to open a cursor
  ********************************************************************
  FUNCTION OPEN( toCursor, tlPopulate )
    LOCAL lnSelect
    *** Save work area
    lnSelect = SELECT()
    *** Use Query to open cursor
    *** If tlPopulate is set, get all records ( '' ), otherwise use 0 = 1
    THIS.QUERY( toCursor, IIF( EMPTY( tlPopulate ), '0 = 1', '' ) )
    *** Restore work area
    SELECT ( lnSelect )
  ENDFUNC

  ********************************************************************
  *** [E] xBhvMgr::QUERY(): Create and run an SQL Query
  ********************************************************************
  FUNCTION QUERY( toCursor, tcWhere )
    THIS.NotImp( toCursor, PROGRAM() )
  ENDFUNC

  ********************************************************************
  *** [E] xBhvMgr::REQUERY(): Repeat the last Query executed for cursor
  ********************************************************************
  FUNCTION REQUERY( toCursor )
    LOCAL lcWhere
    lcWhere = toCursor.cLastWhere
    THIS.QUERY( toCursor, lcWhere )
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::SENDDELETE(): Form where clause and execute a delete
  ********************************************************************
  PROTECTED FUNCTION SendDelete( toCursor )
    LOCAL lcUpdWhere, llOk
    *** Get update WHERE clause
    lcUpdWhere = THIS.GetLocateExp( toCursor )
    llOk = THIS.DELETE( ALLTRIM( toCursor.cUpdTable ), lcUpdWhere )
    *** Return success / failure
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::SENDINSERT(): Form and execute an Insert
  ********************************************************************
  PROTECTED FUNCTION SendInsert( toCursor )
    LOCAL lcAlias, lcType, lcKeyField, lnFieldCount, llHasFields
    LOCAL lnCnt, lcField, lcFldStr, luValue, lcValStr, lcFType
    lcAlias = toCursor.ALIAS
    lcType  = RIGHT( toCursor.cConnName, 3)
    lcKeyField = UPPER( ALLTRIM( toCursor.cKeyField ))
    STORE "" TO lcFldStr, lcValStr
    *** Find out how many updateable fields we have defined
    lnFieldCount = THIS.FieldCount( toCursor.cUpdFields )
    IF lnFieldCount == 0
      *** Default to All fields in the Cursor
      lnFieldCount = FCOUNT( lcAlias )
      llHasFields = .F.
    ELSE
      *** Specific fields have been defined
      llHasFields = .T.
    ENDIF
    FOR lnCnt = 1 TO lnFieldCount
      *** If we have fields defined use them
      IF llHasFields
        lcField = UPPER( ALLTRIM( THIS.GetItem( toCursor.cUpdFields, lnCnt ) ))
      ELSE
        lcField = UPPER( ALLTRIM( FIELD( lnCnt, lcAlias ) ))
      ENDIF

      IF lcField == lcKeyField
        *** Never update the key field on the back end!
        LOOP
      ENDIF
      *** If field has changed build an update string
      IF INLIST( NVL( GETFLDSTATE( lcField, lcAlias ), 0 ), 2, 4 )  && field has changed
        lcFldStr = lcFldStr + IIF( !EMPTY( lcFldStr ), ", ", "" ) + lcField
        luValue  = EVAL( lcAlias + '.' + lcField )
        lcFType  = TYPE( "EVAL( lcAlias + '.' + lcField )" )
        *** If it's a string we need to right trim it and convert all ' to ''
        lcValStr = lcValStr + IIF( !EMPTY( lcValStr ), ", ", "" ) + THIS.PARENT.ExpToStr( luValue, lcFType, lcType )
      ENDIF
    NEXT
    *** If we have an insert string - apply it
    IF ! EMPTY( lcFldStr )
      llOk = THIS.INSERT( toCursor.cUpdTable, lcFldStr, lcValStr )
    ELSE
      llOk = .T.
    ENDIF
    *** Return success / failure
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::SENDUPDATE(): Form and execute an Update
  ********************************************************************
  PROTECTED FUNCTION SENDUPDATE( toCursor )
    LOCAL lcUpdStr, lnCnt, luValue, llOk, lnFieldCount, llHasFields, lcFType
    LOCAL lcAlias, lcUpdTable, lnSpacePos, lcDomain, lcType, lcKeyField
    lcKeyField = UPPER( ALLTRIM( toCursor.cKeyField ))
    lcType  = RIGHT( toCursor.cConnName, 3 )
    lcAlias = toCursor.ALIAS
    lcUpdStr = ""
    lnFieldCount = THIS.FieldCount( toCursor.cUpdFields )
    IF lnFieldCount == 0
      lnFieldCount = FCOUNT( lcAlias )
      llHasFields = .F.
    ELSE
      llHasFields = .T.
    ENDIF
    FOR lnCnt = 1 TO lnFieldCount
      *** If we have fields defined use them
      IF llHasFields
        lcField = UPPER( ALLTRIM( THIS.GetItem( toCursor.cUpdFields, lnCnt ) ))
      ELSE
        lcField = UPPER( ALLTRIM( FIELD( lnCnt, lcAlias ) ))
      ENDIF

      IF lcField == lcKeyField
        *** Never update the key field on the back end!
        LOOP
      ENDIF
      *** If field has changed build an update string
      IF INLIST( NVL( GETFLDSTATE( lcField, lcAlias ), 0 ), 2, 4 )  && field has changed
        lcUpdStr = lcUpdStr + IIF( !EMPTY( lcUpdStr ), ", ", "" ) + lcField + ' = '
        luValue  = EVAL( lcAlias + '.' + lcField )
        lcFType  = TYPE( "EVAL( lcAlias + '.' + lcField )" )
        *** Add value to string
        lcUpdStr = lcUpdStr + THIS.PARENT.ExpToStr( luValue, lcFType ,lcType )
      ENDIF
    NEXT
    *** If we have an update string - apply it
    IF ! EMPTY( lcUpdStr )
      *** Get update WHERE clause
      lcUpdWhere = THIS.GetLocateExp( toCursor )

      *** If the update table has an alias, remove it.
      lnSpacePos = AT_C( SPACE( 1 ), toCursor.cUpdTable )
      lcUpdTable = IIF( lnSpacePos > 0, LEFT( toCursor.cUpdTable, lnSpacePos - 1 ), ;
        toCursor.cUpdTable )
      *** Call the Update Process
      llOk = THIS.UPDATE( lcUpdTable, lcUpdStr, lcUpdWhere )
    ELSE
      llOk = .T.
    ENDIF
    *** Return success / failure
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::SETBUFFERING(): Set buffer mode for cursor
  ********************************************************************
  FUNCTION SetBuffering( toCursor )
    *** Need Multilocks to be set before changing buffer mode
    IF ! 'ON' $ SET("MULTILOCKS")
      SET MULTILOCKS ON
    ENDIF
    RETURN .T.
  ENDFUNC

  ********************************************************************
  *** [E] xBhvMgr::TABLEUPDATE(): Initiate the Database Update Process
  ********************************************************************
  FUNCTION TABLEUPDATE( toCursor, tlAll, tlForce )
    THIS.NotImp( toCursor, PROGRAM() )
  ENDFUNC

  ********************************************************************
  *** [P] xBhvMgr::UPDATE(): Update a record in the back end table
  ********************************************************************
  FUNCTION UPDATE( tcTable, tcSetString, tcWhereClause )
    LOCAL lcSqlStr, llOk

    *** Build update string
    lcSqlStr = "UPDATE " + tcTable + " SET " + tcSetString
    lcSqlStr = lcSqlStr + IIF( !EMPTY( tcWhereClause ), " WHERE " + tcWhereClause, "" )

    *** Execute Update string and force Override of Connection Exec Method
    llOk = THIS.PARENT.EXEC( 0, lcSqlStr, '', .F., .T. )

    *** Report Status in Dev Mode
    ASSERT llOk MESSAGE "Unable to Update Table " + tcTable

    *** Return success/failure
    RETURN llOk
  ENDFUNC

  ****************************************************************
  *** xBhvMgr::VALTOSTR() VALTOSTR(): Convert value into formatted string equivalent
  ****************************************************************
  FUNCTION ValToStr( tuExp, tcType, tcBHV )
    LOCAL luRetVal
    luRetVal = THIS.PARENT.ExpToStr( tuExp, tcType, tcBHV )
    RETURN luRetVal
  ENDFUNC

  ****************************************************************
  *** xBhvMgr::NotImp() Default Behavior when all else fails
  ****************************************************************
  FUNCTION NotImp( toCursor, tcMethod )  && called for no default action
    THIS.LOGERROR( 9007, tcMethod, LOWER( PROGRAM() ) )
    RETURN .F.
  ENDFUNC
ENDDEFINE

********************************************************************
*** xBhvSql -- Standard SQL Back End Behaviour Class ***********
********************************************************************
DEFINE CLASS xBhvSql AS xBhvMgr

  ********************************************************************
  *** [P] xBhvSQL::GETLOCATEEXP(): Construct expression for SQL WHERE clause
  ********************************************************************
  PROTECTED FUNCTION GetLocateExp( toCursor )
    LOCAL lcKeys, lnCnt, lcLocStr, lcKey, lcAlias
    *** Put alias in local for quicker access
    lcAlias = toCursor.ALIAS
    *** Only build string if there's a current record
    IF ! EOF( lcAlias )
      lcKeys = UPPER( ALLTRIM( toCursor.cKeyField ) )
      lcLocStr = ''
      FOR lnCnt = 1 TO OCCURS( ',', lcKeys ) + 1
        lcKey = THIS.GetItem( lcKeys, lnCnt )
        lcLocStr = IIF( lnCnt>1, lcLocStr + ' AND ', lcLocStr )
        lcLocStr = lcLocStr + lcKey + ' = ' + THIS.PARENT.ExpToStr( EVAL(lcAlias + '.' + lcKey), ,'SQL' )
      NEXT
    ELSE
      *** No record selected - return ''
      lcLocStr = ''
    ENDIF
    *** Return string to be used for Locate
    RETURN lcLocStr
  ENDFUNC

  ********************************************************************
  *** [E] xBhvSQL::QUERY(): Build SQL for Query
  ********************************************************************
  FUNCTION QUERY( toCursor, tcWhere, tlNoSafeSelect, tcTgtCursor )

    LOCAL lcWhere, lcCursor, lcSql, llRetVal, lcActWhere
    lcWhere = IIF( TYPE( 'tcWhere' ) = 'C', tcWhere, '' )
    *** If no cursor specified default to alias
    lcCursor = IIF( TYPE( 'tcTgtCursor' ) = 'C', tcTgtCursor, toCursor.ALIAS )
    *** Build SQL String - first we need the Field list
    lcSql = ''
    lcSql = lcSql + 'SELECT ' + toCursor.cSelFields
    *** Next we need the Tables - need to support ANSI92 as well, so
    *****************************************************************
    **
    *** NOTE: This code is set up to tanslate VFP Logical Values into
    *** Numeric Values .T. => 1, .F. => 0
    **
    *****************************************************************
    IF ! EMPTY( toCursor.cPermJoins )
      *** Uses Joins
      lcSql = lcSql + ' FROM ' + STRTRAN( toCursor.cPermJoins, ".T.", "1" )
      lcSql = STRTRAN( lcSql, ".F.", "0" )
    ELSE
      *** Uses Table list and WHERE Clause
      lcSql = lcSql + ' FROM ' + toCursor.cTables
    ENDIF
    *** Add - on perm where clause
    IF !EMPTY( toCursor.cPermWhere )
      lcActWhere = STRTRAN( toCursor.cPermWhere, ".T.", "1" )
      lcActWhere = STRTRAN( lcActWhere, ".F.", "0" )
    ENDIF
    *** Add - on ad hoc clause
    IF !EMPTY( lcWhere )
      lcActWhere = IIF( !EMPTY( lcActWhere ), lcActWhere + ' AND ', '' ) + STRTRAN( lcWhere, ".T.", "1" )
      lcActWhere = STRTRAN( lcActWhere, ".F.", "0" )
    ENDIF
    IF !EMPTY( lcActWhere )
      lcSql = lcSql + ' WHERE ' + lcActWhere
    ENDIF
    *** Apply default Group by clause
    IF !EMPTY( toCursor.cDefGroupBy )
      lcSql = lcSql + ' GROUP BY ' + toCursor.cDefGroupBy
    ENDIF
    *** Apply default order by clause
    IF !EMPTY( toCursor.cDefOrder )
      lcSql = lcSql + ' ORDER BY ' + toCursor.cDefOrder
    ENDIF
    *** Apply Permanent Having clause
    IF !EMPTY( toCursor.cPermHaving )
      lcSql = lcSql + ' HAVING ' + toCursor.cPermHaving
    ENDIF
    *** Store last where issued into cursor property ( used for requery )
    toCursor.cLastWhere = lcWhere
    *** Execute + return status
    llRetVal = THIS.PARENT.EXEC( SET( 'DATASESSION' ), lcSql, lcCursor, tlNoSafeSelect )
    *** Set updatable ( if not select into different cursor )
    IF ALLTRIM( UPPER( lcCursor ) ) = ALLTRIM( UPPER( toCursor.ALIAS ) )
      IF ! THIS.SetBuffering( toCursor )
        THIS.LOGERROR( 9003, 'Cannot make updatable cursor: ' + toCursor.ALIAS, 'Behaviour Manager Query()' )
        llRetVal = .F.
      ENDIF
    ENDIF
    RETURN llRetVal
  ENDFUNC

  ********************************************************************
  *** [P] xBhvSql::SETBUFFERING(): Set buffer mode for cursor
  ********************************************************************
  PROTECTED FUNCTION SetBuffering( toCursor )
    LOCAL llOk, lcAlias

    *** Return if not used, or not updatable
    lcAlias = toCursor.ALIAS
    IF NOT USED( lcAlias ) OR EMPTY( toCursor.lUpdatable )
      RETURN
    ELSE
      *** Call Parent Class to force Multilocks
      DODEFAULT()
    ENDIF

    llOk = .T.
    IF toCursor.lUseTblBuff
      llOk = llOk AND CURSORSETPROP( 'Buffering', 5, lcAlias )  && Set table buffering
    ELSE
      llOk = llOk AND CURSORSETPROP( 'Buffering', 3, lcAlias  )  && Set row buffering
    ENDIF
    *** Always turn off SendUpdates
    llOk = llOk AND CURSORSETPROP( 'SendUpdates', .F., lcAlias  )
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [E] TABLEUPDATE(): Initiate the Database Update Process
  ********************************************************************
  FUNCTION TABLEUPDATE( toCursor )
    LOCAL llOk, lcAlias

    *** Check cursor is updatable -- error if not
    lcAlias = toCursor.ALIAS
    IF ! toCursor.lUpdatable OR ! INLIST( CURSORGETPROP( 'Buffering', lcAlias ), 3, 5 )
      RETURN .F.
    ENDIF

    *** Ensure that SendUpdates is NOT set
    CURSORSETPROP( 'SendUpdates', .F., lcAlias )

    *** Start a Back End Transaction if necessary
    THIS.PARENT.CheckTransLevel()

    *** Send updates to back - end server
    llOk = THIS.ChkUpdType( toCursor )

    *** Return success / failure
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [E] xBhvSql::VALTOSTR(): Convert value into formatted string equivalent
  ********************************************************************
  FUNCTION ValToStr( tuExp, tcType )
    LOCAL luRetVal
    luRetVal = THIS.PARENT.ExpToStr( tuExp, tcType, "SQL" )
    RETURN luRetVal
  ENDFUNC

ENDDEFINE

****************************************************************
*** xBhvVFP -- Standard VFP Behaviour Class ********************
****************************************************************
DEFINE CLASS xBhvVFP AS xBhvMgr

  ********************************************************************
  *** [P] xBhvVFP::ADDDBC(): Add DBC Name to Table Names
  ********************************************************************
  PROTECTED FUNCTION AddDBC( toCursor )
    LOCAL lcDBC, lcTables, lcRetVal, lnCnt

    *** Get the Database and Table Properties
    lcDBC    = ALLTRIM( toCursor.DATABASE )
    lcTables = ALLTRIM( toCursor.cTables )
    lcRetVal = ""

    *** Check that we don't have databases specified in the Tables property
    IF OCCURS( "!", lcTables ) > 0
      *** We have a DBC specified already - so just respect it
      lcRetVal = lcTables
    ELSE
      *** Need to add the specified DBC
      FOR lnCnt = 1 TO OCCURS( ",", lcTables ) + 1
        lcCurTable = THIS.GetItem( lcTables, lnCnt )
        IF ISNULL( lcCurTable )
          *** We have reached the end of the string
          EXIT
        ELSE
          *** Add the Database name and add to return string
          lcCurTable = lcDBC + "!" + lcCurTable
          lcRetVal = lcRetVal + IIF( ! EMPTY( lcRetVal ), ",", "" ) + lcCurTable
        ENDIF
      NEXT
    ENDIF
    *** Return the Results
    RETURN lcRetVal
  ENDFUNC

  ********************************************************************
  *** [P] xBhvVFP::GETLOCATEEXP(): Construct expression for SQL WHERE clause
  ********************************************************************
  PROTECTED FUNCTION GetLocateExp( toCursor )
    LOCAL lcKeys, lnCnt, lcLocStr, lcKey, lcAlias
    *** Put alias in local for quicker access
    lcAlias = toCursor.ALIAS
    *** Only build string if there's a current record
    IF ! EOF( lcAlias )
      lcKeys = UPPER( ALLTRIM( toCursor.cKeyField ) )
      lcLocStr = ''
      FOR lnCnt = 1 TO OCCURS( ',', lcKeys ) + 1
        lcKey = THIS.GetItem( lcKeys, lnCnt )
        lcLocStr = IIF( lnCnt>1, lcLocStr + ' AND ', lcLocStr )
        lcLocStr = lcLocStr + lcKey + ' = ' + THIS.PARENT.ExpToStr( OLDVAL( lcAlias + '.' + lcKey, lcAlias ), ,'VFP' )
      NEXT
    ELSE
      *** No record selected - return ''
      lcLocStr = ''
    ENDIF
    *** Return string to be used for Locate
    RETURN lcLocStr
  ENDFUNC

  ********************************************************************
  *** [E] xBhvVFP::QUERY(): Build SQL for Query
  ********************************************************************
  FUNCTION QUERY( toCursor, tcWhere, tlNoSafeSelect, tcTgtCursor )
    LOCAL lcWhere, lcCursor, lcSql, llRetVal, lcActWhere, llOverRide, llIsUpdatable, lcTableList

    lcWhere = IIF( TYPE( 'tcWhere' ) = 'C', tcWhere, '' )
    *** If no cursor specified default to alias
    lcCursor = IIF( TYPE( 'tcTgtCursor' ) = 'C', tcTgtCursor, toCursor.ALIAS )

    *** Build SQL String - first we need the Field list
    lcSql = ''
    lcSql = lcSql + 'SELECT ' + toCursor.cSelFields

    *** Next we need the Tables - need to support ANSI92 as well, so
    IF ! EMPTY( toCursor.cPermJoins )
      *** Uses Joins - Must assume proper syntax here!
      lcSql = lcSql + ' FROM ' + toCursor.cPermJoins
    ELSE
      *** Uses Table list and WHERE Clause, but we must check
      *** to see if the DataBase property has anything in it
      IF ! EMPTY( toCursor.DATABASE )
        *** We have a DBC Specified
        lcTableList = THIS.AddDBC( toCursor )
      ELSE
        *** We Only have Table Names
        lcTableList = ALLTRIM( toCursor.cTables )
      ENDIF
      lcSql = lcSql + ' FROM ' + lcTableList
    ENDIF

    *** Add - on perm where clause
    IF !EMPTY( toCursor.cPermWhere )
      lcActWhere = toCursor.cPermWhere
    ENDIF

    *** Add - on ad hoc clause
    IF !EMPTY( lcWhere )
      lcActWhere = IIF( !EMPTY( lcActWhere ), lcActWhere + ' AND ', '' ) + lcWhere
    ENDIF

    IF !EMPTY( lcActWhere )
      lcSql = lcSql + ' WHERE ' + lcActWhere
    ENDIF

    *** Apply default Group by clause
    IF !EMPTY( toCursor.cDefGroupBy )
      lcSql = lcSql + ' GROUP BY ' + toCursor.cDefGroupBy
    ENDIF

    *** Apply default order by clause
    IF !EMPTY( toCursor.cDefOrder )
      lcSql = lcSql + ' ORDER BY ' + toCursor.cDefOrder
    ENDIF

    *** Apply Permanent Having clause
    IF !EMPTY( toCursor.cPermHaving )
      lcSql = lcSql + ' HAVING ' + toCursor.cPermHaving
    ENDIF

    *** Store last where issued into cursor property ( used for requery )
    toCursor.cLastWhere = lcWhere

    *** Execute + return status
    *** NB: VFP Code needs an extra parameter here to determine whether to make
    *** the resulting cursor updatable or not.  Not needed with back end cursors
    llOverRide = .F.
    llIsUpdatable = toCursor.lUpdatable
    llRetVal = THIS.PARENT.EXEC( SET( 'DATASESSION' ), lcSql, lcCursor, tlNoSafeSelect, llOverRide, llIsUpdatable )

    *** Set updatable ( if not select into different cursor )
    IF ALLTRIM( UPPER( lcCursor ) ) = ALLTRIM( UPPER( toCursor.ALIAS ) )
      IF ! THIS.SetBuffering( toCursor )
        THIS.LOGERROR( 9003, 'Cannot make updatable cursor: ' + toCursor.ALIAS, 'Behaviour Manager Query()' )
      ENDIF
    ENDIF

    RETURN ( USED( lcCursor ) AND RECCOUNT( lcCursor )>0 )
  ENDFUNC

  ********************************************************************
  *** [P] xBhvVFP::SETBUFFERING(): Set buffer mode for cursor
  ********************************************************************
  PROTECTED FUNCTION SetBuffering( toCursor )
    LOCAL lcAlias, llOk
    *** Return if not used, or not updatable
    lcAlias = toCursor.ALIAS
    IF NOT USED( lcAlias ) OR EMPTY( toCursor.lUpdatable )
      RETURN
    ELSE
      *** Call Parent Class to force Multilocks
      DODEFAULT()
    ENDIF

    *** Note work area + select cursor
    llOk = .T.
    *** Always buffer to row level at least
    IF toCursor.lUseTblBuff
      llOk = llOk AND CURSORSETPROP( 'Buffering', 5, lcAlias )
    ELSE
      llOk = llOk AND CURSORSETPROP( 'Buffering', 3, lcAlias )
    ENDIF
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [E] xBhvVFP::TABLEUPDATE(): Initiate the Database Update Process
  ********************************************************************
  FUNCTION TABLEUPDATE( toCursor )
    LOCAL llOk
    *** Check cursor is updatable and buffered -- do nothing if not
    IF NOT toCursor.lUpdatable OR NOT CURSORGETPROP( 'Buffering', toCursor.ALIAS ) > 1
      RETURN .T.
    ENDIF
    WITH THIS
      *** Check for changes - just return .T. if no changes
      IF NOT .IsChanged( toCursor )
        RETURN .T.
      ENDIF
      *** Make sure we have a transaction on the connection
      .PARENT.CheckTransLevel()
      *** Send updates to back - end server
      llOk = .ChkUpdType( toCursor )
    ENDWITH
    *** Return success / failure
    RETURN llOk
  ENDFUNC

  ********************************************************************
  *** [E] xBhvVFP::VALTOSTR(): Convert value into formatted string equivalent
  ********************************************************************
  FUNCTION ValToStr( tuExp, tcType )
    LOCAL luRetVal
    luRetVal = THIS.PARENT.ExpToStr( tuExp, tcType, "VFP" )
    RETURN luRetVal
  ENDFUNC

ENDDEFINE

*/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/
*/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/

********************************************************************
*** xConMgr - Root Connection Manager Class Definition
********************************************************************
DEFINE CLASS xConMgr AS xObjBase
  DIMENSION aBhvs[1]        && Behavior Collection
  nBhvCnt = 0               && Number of behavior Objects
  nHandle = 0               && Handle of ODBC connection
  nTranCount = 0            && Transaction Counter
  PROTECTED nHandle, nTranCount

  ********************************************************************
  *** [E] ADDBEHAVIOR(): Add a behavior object to this connection
  ********************************************************************
  FUNCTION AddBehavior( tcBHV )
    LOCAL llRetVal, lcObjName
    WITH THIS
      *** Connection will accept any behavior ( however naughty ) !
      lcObjName = 'o' + tcBHV
      IF NOT PEMSTATUS( This, lcObjName, 5 )
        *** Add behavior
        .ADDOBJECT( lcObjName, 'x' + tcBHV )
        llRetVal = ( TYPE( 'This.o' + tcBHV ) = 'O' )
        *** Log Fatal Error if Failed
        IF ! llRetVal
          .LOGERROR( 9007, 'Behavior Object not available: ' + tcBHV, "Connection Manager AddBehavior")
        ELSE
          THIS.UpdCollection( 'ADD', tcBHV )
        ENDIF
        *** Return success/failure
      ELSE
        llRetVal = .T.
      ENDIF
      RETURN llRetVal
    ENDWITH
  ENDFUNC

  *******************************************************************
  *** CHECKTRANSLEVEL(): Check for open transaction on back
  *** end connection, start one if none there
  *******************************************************************
  FUNCTION CheckTransLevel
    RETURN .T.  && default to always succeed
  ENDFUNC

  ********************************************************************
  *** [E] COMMIT(): Commit open transaction
  ********************************************************************
  FUNCTION Commit( tnTxnLevel )
    RETURN .T.  && default to always succeed
  ENDFUNC

  ********************************************************************
  *** [E] CONNECT(): Connect to the specified data source
  ********************************************************************
  FUNCTION CONNECT( tcODBCSource, tcUsername, tcPassword )
    RETURN .T.  && default to always succeed
  ENDFUNC

  ********************************************************************
  *** [E] DESTROY(): Disconnect when object is released
  ********************************************************************
  FUNCTION DESTROY()
    *** Disconnect on object Destroy
    THIS.Disconnect()
  ENDFUNC

  ********************************************************************
  *** [P] DISCONNECT(): Release connection
  ********************************************************************
  FUNCTION Disconnect()
    RETURN .T.  && default to always succeed
  ENDFUNC

  ********************************************************************
  *** [P] DOEXEC(): Execute SQL Statement
  ********************************************************************
  PROTECTED FUNCTION DoExec( tcExecStr, tcCursor )
  ENDFUNC

  ********************************************************************
  *** [E] EXEC(): Prepare SQL Statement for execution
  ********************************************************************
  FUNCTION EXEC( tnDS, tcExecStr, tcCursor, tlNoSafeSelect )
    RETURN .F.  && default to always fail
  ENDFUNC

  ********************************************************************
  *** [E] EXECPROC(): Execute the specified Proc or Function
  ********************************************************************
  FUNCTION ExecProc( tnDS, tcProcName, tuParams )
    *** Execute a Stored Procedure
    RETURN .F.  && default to always fail
  ENDFUNC

  ********************************************************************
  *** [E] EXPTOSTR(): Convert expression to string
  ********************************************************************
  FUNCTION ExpToStr( tuExp, tcType, tcBHV )
    *** Default to Empty String
    RETURN ""
  ENDFUNC

  ********************************************************************
  *** [E] GETERRORS(): Extracts errors from the aErrors collection,
  *** packages them up in an object, and returns it
  ********************************************************************
  FUNCTION GetErrors()
    LOCAL loParamObj, lnCnt, loBhv
    *** Get the errors for each behavior object
    FOR lnCnt = 1 TO THIS.nBhvCnt
      *** Get the object Ref
      loBhv = EVALUATE( "This." + THIS.aBhvs[ lnCnt ] )
      IF VARTYPE( loBhv ) = "O"
        *** Add it's errors to this collection
        THIS.AddErrors( loBhv )
      ENDIF
    NEXT
    *** And then do the Root Class Behavior
    loParamObj = DODEFAULT()
    RETURN loParamObj
  ENDFUNC

  ********************************************************************
  *** [E] GETHANDLE(): Return connection handle (0 if not cnnected)
  ********************************************************************
  FUNCTION GetHandle()
    *** Return handle if connected, 0 if not
    RETURN IIF( THIS.IsConnected(), THIS.nHandle, 0 )
  ENDFUNC

  ****************************************************************
  *** [E] GETLASTID(): Get the last PK Value issued
  ****************************************************************
  FUNCTION GetLastID( tnDS, tcAliasToCheck )
    *** Default to Fail
    RETURN 0
  ENDFUNC

  ********************************************************************
  *** [P] ISBUSY(): Returns availability of specified connection
  ********************************************************************
  PROTECTED FUNCTION IsBusy( tnHandle )
    RETURN .F.  && default to always succeed
  ENDFUNC

  ********************************************************************
  *** [E] ISCONNECTED(): Returns whether connection is active
  ********************************************************************
  FUNCTION IsConnected()
    RETURN .T.  && default to always succeed
  ENDFUNC

  ********************************************************************
  *** [E] ROLLBACK(): Rollback an uncommitted transaction
  ********************************************************************
  FUNCTION ROLLBACK( tnTxnLevel )
    RETURN .T.  && default to always succeed
  ENDFUNC

  ********************************************************************
  *** [P] UPDCOLLECTION(): Update the Behavior collection
  ********************************************************************
  PROTECTED FUNCTION UpdCollection( tcAction, tcBhvName )
    LOCAL lcBhv, lcBhvCnt
    WITH THIS
      lcBhv = 'o' + LOWER( ALLTRIM( tcBhvName ) )
      IF tcAction = "ADD"
        *** How many behaviors currently?
        lcBhvCnt = ALEN( .aBhvs, 1 )
        *** If only one row, is it actually a behavior?
        IF lcBhvCnt = 1 AND EMPTY( .aBhvs[ 1, 1 ] )
          *** Nope - set behavior count = 1
          lcBhvCnt = 1
        ELSE
          *** We already have at least one behavior
          lcBhvCnt = lcBhvCnt + 1
        ENDIF
        DIMENSION .aBhvs[ lcBhvCnt, 1 ]
        .aBhvs[ lcBhvCnt, 1 ] = lcBhv
        .nBhvCnt = .nBhvCnt + 1
      ELSE
        *** Remove the behavior. Find which row it is
        lcBhvCnt = ASCAN( .aBhvs, lcBhv, 1, -1, -1, 9)
        *** And Delete it
        ADEL( .aBhvs, lcBhvCnt )
        *** Decrement the counter
        .nBhvCnt = IIF(.nBhvCnt > 1, .nBhvCnt - 1, 1 )
        *** And re-dimension the Array
        DIMENSION .aBhvs( .nBhvCnt )
      ENDIF
    ENDWITH
    RETURN
  ENDFUNC

  ********************************************************************
  *** [P] ZAPCURSOR(): Clearsa cursor by zapping it
  ********************************************************************
  PROTECTED FUNCTION ZapCursor( tcCursor )
    LOCAL lcOldSafety
    *** If uncommited changes in cursor - clear - out
    IF CURSORGETPROP( 'BUFFERING', tcCursor ) > 1
      TABLEREVERT( .T., tcCursor )
    ENDIF
    *** If cursor isn't table set updatable off
    IF CURSORGETPROP( 'SourceType', tcCursor ) # 3
      *** Always set SendUpdates Off as a matter of course
      CURSORSETPROP( 'SendUpdates', .F., tcCursor )
    ENDIF
    SET MULTILOCKS ON
    *** Force to Row Buffering to allow ZAP.
    CURSORSETPROP( 'Buffering', 3, tcCursor )
    *** Zap records in cursor
    lcOldSafety = SET( 'SAFETY' )
    SET SAFETY OFF
    ZAP IN ( tcCursor )
    IF lcOldSafety = 'ON'
      SET SAFETY ON
    ENDIF
  ENDFUNC

ENDDEFINE

********************************************************************
*** xConSQL - Connection Manager for ODBC Connection to SQL Back End
********************************************************************
DEFINE CLASS xConSQL AS xConMgr

  *******************************************************************
  *** xConSQL::CheckTransLevel() Check for open transaction on back
  *** end connection, start one if none there
  *******************************************************************
  FUNCTION CheckTransLevel
    LOCAL lcSqlStr, lnRetVal, lnSelect
    WITH THIS
      *** Check the connection status - if automatic, do nothing
      IF SQLGETPROP( This.nHandle, "Transactions" ) # 2
        *** Not in manual mode anyway
        RETURN
      ENDIF
      lnRetVal = 0
      lnSelect = SELECT()
      *** Check the transaction status and start one off if needed
      lcSqlStr = "SELECT @@TRANCOUNT nTrans"
      lnRetVal = THIS.EXEC( SET("DATASESSION"), lcSqlStr, '', .T., .T. )
      IF sqlresult.nTrans = 0
        *** We need one here
        lcSqlStr = 'BEGIN TRANSACTION'
        lnRetVal = THIS.EXEC( SET("DATASESSION"), lcSqlStr, '', .T., .T. )
        .nTranCount = 1
      ELSE
        .nTranCount = sqlresult.ntrans
      ENDIF
      IF USED( 'sqlresult' )
        USE IN sqlresult
      ENDIF  
      SELECT (lnSelect)
      RETURN
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [E] COMMIT(): Commit open transaction
  ********************************************************************
  FUNCTION Commit()
    LOCAL lcSqlStr, lnRetVal, lnSelect
    lnRetVal = 0
    lnSelect = SELECT()
    *** Check the transaction status and kill them off 
    WITH THIS
      lcSqlStr = "SELECT @@TRANCOUNT nTrans"
      DO WHILE .T.
        lnRetVal = .EXEC( SET("DATASESSION"), lcSqlStr, '', .T., .T. )
        IF sqlresult.nTrans > 0
          *** We need to close one here
          SQLCOMMIT( This.nHandle )
        ELSE
          .nTranCount = 0
          EXIT
        ENDIF
      ENDDO
      SELECT (lnSelect)
      RETURN
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [E] CONNECT(): Connect to the specified data source
  ********************************************************************
  FUNCTION CONNECT( tcUid, tcPwd, tcDSN )
    LOCAL lnRetVal
    *** Attempt to connect to the specified DSN
    lnRetVal = SQLCONNECT( tcDSN, tcUid, IIF( !EMPTY( tcPwd), tcPwd, '' ) )
    WITH THIS
      *** Set handle on successful connect or 0 if not
      .nHandle = IIF( lnRetVal>0, lnRetVal, 0 )
      *** If connected -- set connection properties
      IF .IsConnected()
        IF ! .SetConnParms( tcDSN )  && unable to set connection parameters
          .Disconnect()
        ENDIF
      ENDIF
      *** Return .T. if connect ok or .F. if not
      RETURN .IsConnected()
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [E] DISCONNECT(): Release connection
  ********************************************************************
  FUNCTION Disconnect()
    WITH THIS
      *** Disconnect if valid handle
      IF !EMPTY( .nHandle )
        *** Ensure rollback on exit
        .ROLLBACK()
        *** Do disconnect ( TYPE() is used so no error if handle invalid )
        IF TYPE( 'SQLDISCONNECT( ' + STR( .nHandle, 10 ) + ' )' ) = 'L'
          *** Dummy Really, nothing needed here anyway
        ENDIF
      ENDIF
      *** Reset handle to 0 to indicate no connection
      .nHandle = 0
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [P] DOEXEC(): Execute SQL Statement
  ********************************************************************
  PROTECTED FUNCTION DoExec( tcExecStr, tcCursor, tlOverRide )
    LOCAL lnRetVal, lcCursor

    IF TYPE( 'glCopySQL' ) = "L"
      IF glCopySQL AND ! ' TRANSACTION' $ UPPER( tcExecStr )
        _CLIPTEXT = tcExecStr
      ENDIF
    ENDIF
    IF ! EMPTY( tlOverRide )
      *** No return Value Needed
      lnRetVal = SQLEXEC( THIS.nHandle, tcExecStr )
    ELSE
      *** If no cursor specified, use SQLRESULT
      lcCursor = IIF( !EMPTY( tcCursor ), ALLTRIM( tcCursor ), 'sqlresult' )
      *** Execute SQL Statement
      lnRetVal = SQLEXEC( THIS.nHandle, tcExecStr, lcCursor )
    ENDIF
    IF lnRetVal < 0
      THIS.ODBCErrNice()
    ENDIF
    *** Return success status
    RETURN lnRetVal
  ENDFUNC

  ********************************************************************
  *** [E] EXEC(): Prepare SQL Statement for execution
  ********************************************************************
  FUNCTION EXEC( tnDS, tcExecStr, tcCursor, tlNoSafeSelect, tlOverRide )
    LOCAL lnOldDS, lcCursor, lnRetVal, lcTmpCur

    *** If OverRide flag is set, just call DoExec with the ExecString
    IF ! EMPTY( tlOverRide )
      RETURN ( THIS.DoExec( tcExecStr, '', tlOverRide ) = 1 )
    ENDIF

    *** Save DS + set to passed DS
    lnOldDS = SET( 'DATASESSION' )
    SET DATASESSION TO ( tnDS )

    *** If no cursor specified, use SQLRESULT
    lcCursor = IIF( !EMPTY( tcCursor ), ALLTRIM( tcCursor ), 'sqlresult' )

    *** If cursor exists -- zap
    IF USED( lcCursor )
      THIS.ZapCursor( lcCursor )
    ENDIF

    *** If no safe select specified or no cursor - select straight into cursor
    IF !USED( lcCursor ) OR !EMPTY( tlNoSafeSelect )
      *** Select straight into cursor
      lnRetVal = THIS.DoExec( tcExecStr, lcCursor )
    ELSE
      *** Execute select into temp cursor
      lcTmpCur = SYS( 2015 )
      lnRetVal = THIS.DoExec( tcExecStr, lcTmpCur )
      SELECT ( lcCursor )
      IF lnRetVal > 0 AND USED( lcTmpCur )
        *** Append new records in + close temp cursor
        APPEND FROM DBF( lcTmpCur )
        GO TOP
        USE IN ( lcTmpCur )
      ENDIF
    ENDIF

    *** Restore data session
    SET DATASESSION TO ( lnOldDS )

    *** Return status
    RETURN ( lnRetVal > 0 )
  ENDFUNC

  ********************************************************************
  *** [E] EXECPROC(): Execute the specified Proc or Function
  ********************************************************************
  FUNCTION ExecProc( tnDS, tcProcName, tuParams )
    LOCAL lcSql, lcParams
    *** Construct the SQL String to execute a Stored Procedure
    lcSql = 'EXECUTE ' + ALLTRIM( tcProcName )
    *** Check that we got something in the parameters
    IF VARTYPE( tuParams ) # 'L' AND ! EMPTY( tuParams )
      lcParams = ALLTRIM( TRANSFORM( tuParams ) )
      lcSql = lcSql + " " + lcParams
    ENDIF
    *** Execute the preocedure directly
    lnRes = THIS.DoExec( lcSql, "", .T. )
    IF lnRes < 0
      *** Something failed
      THIS.LOGERROR( 9004, 'Error Executing Stored Procedure ' + lcSql , "Connection Manager ExecProc()" )
    ENDIF
    *** Return the result
    RETURN lnRes
  ENDFUNC

  ********************************************************************
  *** [E] EXPTOSTR(): Convert expression to string
  ********************************************************************
  FUNCTION ExpToStr( tuExp, tcType, tcBHV )
    LOCAL lcMethodToDo, lcRetVal
    lcMethodName = "ValTo" + ALLTRIM( tcBHV ) + "Str"
    IF ! PEMSTATUS( THIS.PARENT, lcMethodName, 5 )
      THIS.LOGERROR( 9004, 'Cannot Find Data Manager ' + lcMethodName, "Connection Manager ExpToStr()" )
      RETURN NULL
    ENDIF
    lcMethodToDo = "This.Parent.ValTo" + ALLTRIM( tcBHV ) + "Str( tuExp, tcType )"
    lcRetVal = &lcMethodToDo
    RETURN lcRetVal
  ENDFUNC

  ****************************************************************
  *** [E] GETLASTID(): Returns the last ID value generated on this connection
  ****************************************************************
  FUNCTION GetLastID( tnDS, tcAliasToCheck  )
    *** Returns the last Identity Value generated on this connection
    LOCAL lnRetVal, lnSelect, llOk, lnSess
    lnSelect = SELECT()
    lnSess = IIF( EMPTY(tnDS), SET( 'DATASESSION' ), tnDS )
    *** Get the last Identity Column Value used
    llOk = THIS.EXEC( lnSess, "SELECT @@IDENTITY AS ident", "_csrlastid", .T. )
    *** If all OK read the value (may be NULL!)
    lnRetVal = IIF( llOk, NVL(_csrlastid.ident, 0), 0 )
    IF USED( "_csrlastid" )
      USE IN _csrlastid
    ENDIF
    SELECT (lnSelect)
    RETURN lnRetVal
  ENDFUNC

  ********************************************************************
  *** [P] ISBUSY(): Returns availability of specified connection
  ********************************************************************
  PROTECTED FUNCTION IsBusy( tnHandle )
    LOCAL llRetVal, lnHandle
    lnHandle = IIF( VARTYPE( tnHandle ) = "N", tnHandle, THIS.nHandle )
    llRetVal = SQLGETPROP( lnHandle, "ConnectBusy" )
    RETURN llRetVal
  ENDFUNC

  ********************************************************************
  *** [E] ISCONNECTED(): Returns whether connection is active
  ********************************************************************
  FUNCTION IsConnected( tnHandle )
    PRIVATE plNoError
    LOCAL llRetVal, lnHandle, lcOldErr
    *** If no handle passed, used current property
    lnHandle = IIF( VARTYPE( tnHandle ) = "N", tnHandle, THIS.nHandle )
    IF lnHandle < 1
      *** We definitely don't have a connection
      RETURN .F.
    ENDIF
    *** Temporarily Disable Error Handling
    lcOldErr = ON( "ERROR" )
    ON ERROR plNoError = .F.
    plNoError = .T.
    *** Check the connection - if handle is invalid will get an error
    *** And return value is an empty string
    llRetVal = THIS.IsBusy( lnHandle )
    *** Restore the Error Handler
    ON ERROR &lcOldErr
    *** Return the result
    RETURN plNoError
  ENDFUNC

  ****************************************************************
  *** xConSQL::ODBCErrNice() *************************************
  ****************************************************************
  FUNCTION ODBCErrNice( tnDS )
    LOCAL lnErrNum, lcErrTxt
    LOCAL ARRAY laError[ 1 ]
    *** Get error details
    = AERROR( laError )
    lnErrNum = laError[ 1, 1 ]
    lcErrTxt = laError[ 1, 2 ]
    THIS.LOGERROR( lnErrNum, lcErrTxt, 'Connectivity Error' )
  ENDFUNC

  ********************************************************************
  *** [E] ROLLBACK(): Rollback an uncommitted transaction
  ********************************************************************
  FUNCTION ROLLBACK()
    LOCAL lcSqlStr, llRetVal
    WITH THIS
      *** If no transaction, do nothing
      IF .nTranCount < 1
        RETURN .T.
      ENDIF
      *** SQL Server cannot nest transactions, one Commit/Rollback
      *** Closes ALL open transactions. Rollback cannot fail!
      SQLROLLBACK( This.nHandle )
      *** Reduce the Transaction Counter
      .nTranCount = 0
      RETURN llRetVal
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [P] SETCONNPARMS(): Set connection parameters
  ********************************************************************
  PROTECTED FUNCTION SetConnParms( tcDSN )
    LOCAL ARRAY laProps[1,2]
    LOCAL llOk, lnHandle, lnCnt, lcPropName, luVal
    *** Get handle for better performance
    lnHandle = THIS.nHandle
    *** Get the connection setting properties from the table
    SELECT cConProp, cConPVal ;
      FROM conprops ;
      WHERE UPPER( ALLTRIM( cConDSN )) = UPPER( ALLTRIM( tcDSN ));
      INTO ARRAY laProps
    *** If we got anything, set the connection
    IF _TALLY > 0
      llOk = .T.
      FOR lnCnt = 1 TO ALEN( laProps, 1)
        lcPropName = ALLTRIM( laProps[ lnCnt, 1 ] )
        lcVal = ALLTRIM( laProps[ lnCnt, 2 ])
        IF ISALPHA( lcVal )
          *** Must be a logical Setting Either T/F
          luVal = IIF( UPPER( LEFT( lcVal, 1) ) = "T", .T., .F. )
        ELSE
          *** It's numeric
          luVal = INT( VAL( lcVal ) )
        ENDIF
        llOk = ( SQLSETPROP( lnHandle, lcPropName, luVal ) = 1 )  AND llOk
      NEXT
    ELSE
      *** Drop this connection!
      llOk = .F.
    ENDIF

    *** But close the table first
    IF USED( 'conprops' )
      USE IN conprops
    ENDIF

    *** Return the status
    RETURN llOk
  ENDFUNC
ENDDEFINE

********************************************************************
*** xConVFP - Connection Manager for Connecting to VFP Database
********************************************************************
DEFINE CLASS xConVFP AS xConMgr
  DIMENSION aDBC[1]        && Array of Databases Opened by this connection
  nHandle = 0              && Default to No Connection
  PROTECTED nHandle

  ****************************************************************
  *** xConVFP::AddDBC() ******************************************
  ****************************************************************
  FUNCTION AddDBC( tcDBCToAdd, tcDBCName )
    LOCAL lnDBC

    *** Check the collection of databases
    lnDBC = ALEN( THIS.aDBC, 1 )
    *** Add a new item if needed
    IF lnDBC # 1 OR ! EMPTY( THIS.aDBC[1] )
      lnDBC = lnDBC + 1
    ENDIF
    DIMENSION THIS.aDBC[ lnDBC, 2 ]
    *** Add the DBC Name
    THIS.aDBC[ lnDBC, 1 ] = tcDBCName
    THIS.aDBC[ lnDBC, 2 ] = tcDBCToAdd
  ENDFUNC

  *******************************************************************
  *** xConVFP::CHECKTRANSLEVEL() Check for open transaction on back
  *** end connection, start one if not one in place already
  *******************************************************************
  FUNCTION CheckTransLevel
    LOCAL lcSqlStr, lnRetVal
    lnRetVal = 0
    *** Check the transaction status and start one off if needed
    WITH THIS
      IF .nTranCount = 0
        *** Need a back end transaction
        BEGIN TRANSACTION
        .nTranCount = .nTranCount + 1
      ENDIF
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [E] CONNECT(): Connect to the specified data source
  ********************************************************************
  FUNCTION CONNECT( tcUsername, tcPassword, tcDatabase )
    LOCAL lcDBCName, lcDBCFile, lcDBCToAdd
    *** Check we got a Database Name
    IF VARTYPE( tcDatabase ) # 'C' OR EMPTY( tcDatabase )
      ASSERT .F. MESSAGE 'Must pass a database name in position 3 of the parameter list'
      RETURN .F.
    ENDIF
    *** Get Database Name and File Name
    lcDBCName = JUSTSTEM( tcDatabase )
    lcDBCFile = FORCEEXT( tcDatabase, 'DBC' )
    *** If database is not already in use, open it
    *** Go to the default data session first
    lnDSID = SET("DATASESSION")
    SET DATASESSION TO 1
    IF ! DBUSED( lcDBCName )
      OPEN DATABASE ( lcDBCFile )
      lcDBCToAdd = JUSTPATH( DBC() )
      SET PATH TO ( SET("PATH") + ';' + lcDBCToAdd )
      THIS.AddDBC( lcDBCToAdd, lcDBCName )
    ENDIF
    *** Restore DataSession
    SET DATASESSION TO (lnDSID)
    RETURN THIS.IsConnected()

  ENDFUNC

  ********************************************************************
  *** [P] DISCONNECT(): Release connection
  ********************************************************************
  FUNCTION Disconnect()
    LOCAL ARRAY laDBC[1]
    LOCAL lnDBC, lcPath, lnCnt, lcDBC
    WITH THIS
      *** Disconnect if valid handle
      IF !EMPTY( .nHandle )
        *** Ensure rollback on exit
        .ROLLBACK()
      ENDIF
      *** Remove Database Names from Path
      lnDBC = 0
      lnDBC = ALEN( THIS.aDBC, 1)
      IF lnDBC > 0
        lcPath = SET( "PATH" )
        FOR lnCnt = 1 TO lnDBC
          *** If still open, make the database current
          lcDBC = THIS.aDBC[ lnCnt, 1 ]
          *** Make sure there really is something there first though!
          IF EMPTY( lcDBC )
            LOOP
          ENDIF
          IF DBUSED( lcDBC )
            SET DATA TO ( lcDBC )
            *** And Close it
            CLOSE DATABASE
          ENDIF
          *** Remove from Path (if in the middle)
          lcDBC =  ';' + THIS.aDBC[ lnCnt, 2 ]
          lcPath = STRTRAN( lcPath, lcDBC )
          *** Remove from end of the path
          lcDBC = CHRTRAN( lcDBC, ';', '' )
          lcPath = STRTRAN( lcPath, lcDBC )
        NEXT
        *** Reset the Path
        SET PATH TO ( lcPath )
      ENDIF
      *** Reset handle to 0 to indicate no connection
      .nHandle = 0
    ENDWITH
  ENDFUNC

  ********************************************************************
  *** [E] COMMIT(): Commit open transaction
  ********************************************************************
  FUNCTION Commit()
    PRIVATE plNoError
    LOCAL lcOldError
    *** If we have no transactions, do nothing
    IF THIS.nTranCount < 1
      RETURN .T.
    ENDIF
    *** Need to disable Error handling here
    lcOldError = ON( "ERROR" )
    plNoError = .T.
    ON ERROR plNoError = .F.
    *** Close Open Transactions (there should only ever be one)
    FOR lnCnt = 1 TO THIS.nTranCount
      END TRANSACTION
    NEXT
    *** Check for error
    IF plNoError
      *** All OK Clear the back end transaction counter
      THIS.nTranCount = 0
    ELSE
      *** Log the error and Rollback
      THIS.LOGERROR( 9018, 'ConVFP', LOWER( PROGRAM()))
      THIS.ROLLBACK()
    ENDIF
    *** Restore Error Handler
    IF NOT EMPTY( lcOldError )
      *** Re-Set
      ON ERROR &lcOldError
    ELSE
      *** Default
      ON ERROR
    ENDIF
    *** Return Status
    RETURN plNoError
  ENDFUNC

  ********************************************************************
  *** [P] DOEXEC(): Execute SQL Statement
  ********************************************************************
  PROTECTED FUNCTION DoExec( tcExecStr, tcCursor, tlOverRide )

    LOCAL lcCursor, lcSql, lnRetVal, lnErrCnt
    IF ! EMPTY( tlOverRide )
      *** This is an override Call
      lcSql = tcExecStr
    ELSE
      *** If no cursor specified, use SQLRESULT
      lcCursor = IIF( !EMPTY( tcCursor ), ALLTRIM( tcCursor ), 'sqlresult' )

      *** Add Target to SQL Statement + suppress filter option
      lcSql = tcExecStr + ' INTO CURSOR ' + lcCursor + ' NOFILTER'
    ENDIF

    *** Save SQL String to Clipboard
    IF TYPE( 'glCopySQL' ) = "L"
      IF glCopySQL AND ! ' TRANSACTION' $ UPPER( tcExecStr )
        _CLIPTEXT = tcExecStr
      ENDIF
    ENDIF

    ************************************************************
    *** Note!
    *** Because VFP does not return a status for executing SQL
    *** We need to check the error count before and after the statement
    *** is executed to see if an error occurred. Of course, it is also a
    *** royal pain because if there is only one row in the error collection
    *** it may be empty anyway
    ************************************************************
    IF ALEN( THIS.aErrors, 1 ) > 0 AND ! EMPTY( THIS.aErrors[1] )
      *** How many errors so far?
      lnErrCnt = ALEN( THIS.aErrors, 1 )
    ELSE
      *** No errors
      lnErrCnt = 0
    ENDIF

    *** Now Execute the SQL
    &lcSql

    *** Now check the error count again and set the return value
    IF ALEN( THIS.aErrors, 1 ) > 0 AND ! EMPTY( THIS.aErrors[1] )
      *** Set to -1 if a new error, otherwise leave at 1
      lnRetVal = IIF( ALEN( THIS.aErrors, 1 ) = lnErrCnt, 1, -1 )
    ELSE
      *** No errors anyway
      lnRetVal = 1
    ENDIF

    *** Return status as 1 or -1 ( to match SQLEXEC return )
    RETURN lnRetVal
  ENDFUNC

  ********************************************************************
  *** [E] EXEC(): Prepare SQL Statement for execution
  ********************************************************************
  FUNCTION EXEC( tnDS, tcExecStr, tcCursor, tlNoSafeSelect, tlOverRide, tlIsUpdatable )
    LOCAL lnOldDS, lcCursor, lnRetVal
    LOCAL ARRAY laData[1], laFlds[1]

    *** If OverRide flag is set, just call DoExec with the ExecString
    IF ! EMPTY( tlOverRide )
      RETURN ( THIS.DoExec( tcExecStr, '', .T. ) = 1 )
    ENDIF

    *** Save DS + set to passed DS
    lnOldDS = SET( 'DATASESSION' )
    SET DATASESSION TO ( tnDS )

    *** If no cursor specified, use SQLRESULT
    lcCursor = IIF( !EMPTY( tcCursor ), ALLTRIM( tcCursor ), 'sqlresult' )

    *** If cursor exists -- zap
    IF USED( lcCursor )
      THIS.ZapCursor( lcCursor )
    ENDIF

    *** If no safe select specified or no cursor - select straight into cursor
    IF ! EMPTY( tlNoSafeSelect )
      *** Select straight into cursor
      lnRetVal = THIS.DoExec( tcExecStr, lcCursor )
    ELSE
      *** Execute select into temp cursor
      LOCAL lcTmpCur
      lcTmpCur = SYS( 2015 )
      lnRetVal = THIS.DoExec( tcExecStr, lcTmpCur )
      *** If we don't actually have the cursor already create it Read/Write
      IF ! USED( lcCursor )
        =AFIELDS( laFlds, lcTmpCur )
        CREATE CURSOR (lcCursor) FROM ARRAY laFlds
      ENDIF
      *** Otherwise, just select it
      SELECT ( lcCursor )
      *** Check for data in temporary cursor
      IF lnRetVal > 0 AND USED( lcTmpCur )
        *** Append new records in + close temp cursor
        APPEND FROM DBF( lcTmpCur )
        GO TOP IN ( lcCursor )
      ENDIF
      USE IN ( lcTmpCur )
    ENDIF

    *** Restore data session
    IF USED( lcCursor )
      SELECT (lcCursor)
    ENDIF
    SET DATASESSION TO ( lnOldDS )

    *** Return status
    RETURN ( lnRetVal = 1 )
  ENDFUNC

  ********************************************************************
  *** [E] EXECPROC(): Execute the specified Proc or Function
  ********************************************************************
  FUNCTION ExecProc( tnDS, tcProcName, tuParams )
    LOCAL lnOldDS, lcProc, luRetVal
    *** Save DS + set to passed DS
    lnOldDS = SET( 'DATASESSION' )
    SET DATASESSION TO ( tnDS )
    *** Get the procedure name
    lcProc = CHRTRAN( tcProcName, '()', '')
    IF PCOUNT() > 2
      *** We had some parameters
      lcProc = lcProc + "(" + tuParams + ")"
    ELSE
      *** No Parameters
      lcProc = lcProc + "()"
    ENDIF
    *** Now execute the procedure
    luRetVal = &lcProc
    *** Restore data session
    SET DATASESSION TO ( lnOldDS )
    *** Tidy Up
    RETURN luRetVal
  ENDFUNC


  ********************************************************************
  *** [E] EXPTOSTR(): Convert expression to string
  ********************************************************************
  FUNCTION ExpToStr( tuExp, tcType, tcBHV )
    LOCAL lcMethodToDo, lcRetVal
    lcMethodName = "ValTo" + ALLTRIM( tcBHV ) + "Str"
    IF ! PEMSTATUS( THIS.PARENT, lcMethodName, 5 )
      THIS.LOGERROR( 9004, 'Cannot find Data Manager ' + lcMethodName, "Connection Manager ExpToStr()" )
      RETURN NULL
    ENDIF
    lcMethodToDo = "This.Parent.ValTo" + ALLTRIM( tcBHV ) + "Str( tuExp, tcType )"
    lcRetVal = &lcMethodToDo
    RETURN lcRetVal
  ENDFUNC

  ****************************************************************
  *** [E] GETLASTID(): Returns the key value from the record in use
  ****************************************************************
  FUNCTION GetLastID( tnDS, toCursor )
    *** Returns the last Identity Value generated for the specified table
    LOCAL lnSelect, lnOldDS
    lnSelect = SELECT()
    lnOldDS = SET( 'DataSession' )
    IF lnOldDS # tnDS
      SET DATASESSION TO ( tnDS )
    ENDIF
    lcKeyField = ALLTRIM( toCursor.cUpdTable ) + '.' + ALLTRIM( toCursor.cKeyField )
    *** Get the last Identity Column Value used
    lnRetVal = &lcKeyField
    IF lnOldDS # tnDS
      SET DATASESSION TO ( lnOldDS )
    ENDIF
    RETURN lnRetVal
  ENDFUNC

  ********************************************************************
  *** [E] ISCONNECTED(): Returns whether connection is active
  ********************************************************************
  FUNCTION IsConnected( tnHandle )
    LOCAL llRetVal, lnHandle, lcOldErr
    *** In VFP here, ignore param, set handle = number of DBCs
    THIS.nHandle = ALEN( THIS.aDBC, 1 )
    *** And simply return .T.
    RETURN .T.
  ENDFUNC

  ********************************************************************
  *** [E] ROLLBACK(): Rollback an uncommitted transaction
  ********************************************************************
  FUNCTION ROLLBACK()
    LOCAL lnCnt
    *** Do we have an open Transaction (other than the front end)
    IF THIS.nTranCount > 1
      *** Close as many as there may be (should only be one!)
      FOR lnCnt = THIS.nTranCount TO 1 STEP -1
        ROLLBACK
      NEXT
      *** And clear the counter
      THIS.nTranCount = 0
    ENDIF
    *** Just Return .T.
    RETURN
  ENDFUNC

ENDDEFINE

*/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/
*/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/**/

DEFINE CLASS xSelAlias AS RELATION
  ******************************************************************
  * Abstract...: Class to Select a specific Work area and restore it
  * ...........: Use as a local object to change work area.
  * ...........: When released will restore the previous work area
  * ...........: Accepts two parameters, the first is the Alias and is mandatory
  * ...........: The second is the name of the table.  If not passed, assumed to be the
  * ...........: same as the Alias you are using.
  * ...........: loSel = CREATEOBJECT( 'xSelAlias', <Alias> [, |<File Name> ] )
  ******************************************************************
  PROTECTED nOldArea
  nOldArea = 0
  PROTECTED lWasOpen
  lWasOpen = .T.
  PROTECTED cAlias
  cAlias = ''

  ****************************************************************
  *** xSelAlias::Init() ******************************************
  ****************************************************************
  PROCEDURE INIT( tcAlias, tcTable )
    LOCAL llRetVal
    *** No Alias Passed - Bail Out
    IF ! VARTYPE( tcAlias ) = "C"
      ASSERT .F. MESSAGE "Must Pass an Alias Name to Work Area Selector"
      RETURN .F.
    ENDIF
    tcAlias = UPPER( ALLTRIM( tcAlias ) )
    IF VARTYPE( tcTable ) # "C" OR EMPTY( tcTable )
      tcTable = tcAlias
    ELSE
      tcTable = UPPER( ALLTRIM( tcTable ) )
    ENDIF
    *** If already in correct work area - do nothing
    IF UPPER( ALLTRIM( ALIAS() ) ) == tcAlias
      RETURN .F.
    ENDIF
    *** If Specified Alis not open - Open it
    IF ! USED( tcAlias )
      USE ( tcTable ) AGAIN IN 0 ALIAS ( tcAlias ) SHARED
      *** And Check!
      llRetVal = USED( tcAlias )
      *** If Forced Open, Note the fact
      IF llRetVal
        THIS.lWasOpen = .F.
      ENDIF
    ELSE
      llRetVal = .T.
    ENDIF
    *** If OK, save current work area and
    *** Now Move to the specified Work Area
    IF llRetVal
      THIS.nOldArea = SELECT()
      SELECT ( tcAlias )
      THIS.cAlias = tcAlias
    ENDIF
    *** Return Status
    RETURN llRetVal
  ENDPROC

  ****************************************************************
  *** xSelAlias::Destroy() ***************************************
  ****************************************************************
  PROCEDURE DESTROY
    WITH THIS
      *** If table opened by this object, close it
      IF ! .lWasOpen
        USE IN ( THIS.cAlias )
      ENDIF
      *** Restore Previous work area
      IF ! EMPTY( .nOldArea )
        SELECT ( .nOldArea )
      ENDIF
    ENDWITH
  ENDPROC

ENDDEFINE
