using System;
using System.IO;
using System.Data;
using System.Xml;
using System.Text;

#if DEBUG
using System.Diagnostics;
#endif

namespace Westwind.WebMonitor.Components
{
	/// <summary>
	/// Summary description for WebMonitorLogParser.
	/// </summary>
	public class WebMonitorLogParser
	{
		/// <summary>
		/// Parses the Error Log and returns a DataTable with the result.
		/// </summary>
		/// <param name="FileName"></param>
		/// <returns>DataTable named wmErrorLog</returns>
		public static DataTable ParseErrorLog(string FileName) 
		{
			if (!File.Exists(FileName))
				return null;

			DataTable dt = new DataTable("wmErrorLog");
			dt.Columns.Add(new DataColumn("SiteName",typeof(string) ));
			dt.Columns.Add(new DataColumn("SiteId",typeof(string) ));
			dt.Columns.Add(new DataColumn("Url",typeof(string) ));
			dt.Columns.Add(new DataColumn("Time",typeof(DateTime) ));
			dt.Columns.Add(new DataColumn("Type",typeof(string)));
			dt.Columns.Add(new DataColumn("Message",typeof(string) ));
			dt.Columns.Add(new DataColumn("SearchFor",typeof(string) ));
			dt.Columns.Add(new DataColumn("lastresponsetext",typeof(string)) );

			XmlTextReader xr = null;
			Stream fs = null;
			try 
			{
				// *** Open stream explicitly so we can use a shared read
				fs = new FileStream(FileName,FileMode.Open,FileAccess.Read,FileShare.Read);
				xr = new XmlTextReader(fs);

				//xr = new XmlTextReader(FileName);
			}
			catch 
			{
				return null; // file not found or available
			}

			DataRow dr = null;
			
			while ( xr.Read() ) 
			{
				
				try 
				{
					if (xr.NodeType == XmlNodeType.Element) 
					{
						if (xr.Name == "event") 
						{
							if (xr.HasAttributes)  
							{
								xr.MoveToAttribute(0);
								dr = dt.NewRow();
								dr["SiteName"] = xr.Value;
								
								if (xr.AttributeCount > 1) 
								{
									xr.MoveToAttribute(1);
									dr["SiteId"] = xr.Value;
								}
							}
						}
						else if (xr.Name == "url")  
						{
							xr.Read(); // read the text
							dr["Url"] = xr.Value;
						}
						else if (xr.Name == "time")
							try 
							{
								xr.Read();
								dr["Time"] = XmlConvert.ToDateTime(xr.Value);
							}
							catch 
							{
								dr["Time"] = Convert.ToDateTime("01/02/1900");
							}
						else if (xr.Name == "message") 
						{
							xr.Read();
							dr["Message"] = xr.Value;
						}
						else if(xr.Name == "type") 
						{
							xr.Read();
							dr["type"] = xr.Value;
						}
						else if(xr.Name == "searchfor")
						{
							xr.Read();
							dr["SearchFor"] = xr.Value;
						}
						else if(xr.Name == "responsetext")
						{
							xr.Read();
							dr["LastResponseText"] = xr.Value;
						}
					}
					if (xr.NodeType == XmlNodeType.EndElement && xr.Name == "event")
						dt.Rows.Add(dr);
				}
				catch(Exception ex)
				{ 
#if DEBUG
					Debug.Assert( false, "Error Parsing XML Error Log Document",ex.Message + "\r\n\r\n" + ex.StackTrace);   // ignore error
#endif
				}
			}

			xr.Close();
			fs.Close();		
	
			return dt;
		}

		/// <summary>
		/// Parses the Error Log and returns a DataTable with the result.
		/// </summary>
		/// <param name="FileName"></param>
		/// <returns></returns>
		public static DataTable ParseDetailLog(string FileName) 
		{
			if (!File.Exists(FileName))
				return null;

			DataTable dt = new DataTable("wmDetailLog");
			dt.Columns.Add(new DataColumn("SiteName",typeof(string) ));
			dt.Columns.Add(new DataColumn("SiteId",typeof(string) ));
			dt.Columns.Add(new DataColumn("Time",typeof(DateTime) ));
			dt.Columns.Add(new DataColumn("Secs",typeof(Decimal) ));
			dt.Columns.Add(new DataColumn("Type",typeof(string) ));
			dt.Columns.Add(new DataColumn("Message",typeof(string) ));

			XmlTextReader xr = null;
			FileStream fs = null;
			try 
			{
				// *** Open stream explicitly so we can use a shared read
				fs = new FileStream(FileName,FileMode.Open,FileAccess.Read,FileShare.Read);
				xr = new XmlTextReader(fs);
			}
			catch 
			{
				return null; // file not found or available
			}

			DataRow dr = null;
			
			while ( xr.Read() ) 
			{
				try 
				{
					if (xr.NodeType == XmlNodeType.Element) 
					{
						if (xr.Name == "event") 
						{
							if (xr.HasAttributes)  
							{
								xr.MoveToAttribute(0);
								dr = dt.NewRow();
								dr["SiteName"] = xr.Value;

								if (xr.AttributeCount > 1) 
								{
									xr.MoveToAttribute(1);
									dr["SiteId"] = xr.Value;
								}

							}
						}
						else if (xr.Name == "type")  
						{
							xr.Read(); // read the text
							dr["type"] = xr.Value;
						}
						else if (xr.Name == "time")
							try 
							{
								xr.Read();
								dr["Time"] = XmlConvert.ToDateTime(xr.Value);
							}
							catch 
							{
								dr["Time"] = Convert.ToDateTime("01/02/1900");
							}
						else if (xr.Name == "secs")
							try 
							{
								xr.Read();
								dr["secs"] = XmlConvert.ToDecimal(xr.Value);
							}
							catch 
							{
								dr["secs"] = 0.00;
							}
						else if (xr.Name == "message") 
						{
							xr.Read();
							dr["Message"] = xr.Value;
						}
					}
					if (xr.NodeType == XmlNodeType.EndElement && xr.Name == "event")
						dt.Rows.Add(dr);

				}
				catch(Exception ex) 
				{
#if DEBUG
					Debug.Assert( false, "Error Parsing XML Detail Log Document",ex.Message + "\r\n\r\n" + ex.StackTrace);  // ignore error
#endif
				}
			}

			xr.Close();
			fs.Close();

			return dt;
		}
	}
}
