using System;
using System.IO;
using System.Text;

using System.Xml;
using System.Xml.Serialization;

using System.Threading;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Westwind.WebMonitor.Components
{
	/// <summary>
	/// Handles individual site access and configuration info
	/// </summary>
	[ComSourceInterfaces( typeof(IWebMonitorSiteEvents) )]
	[ClassInterface(ClassInterfaceType.AutoDual)]
	[ProgId("WebMonitor.WebMonitorSiteList")]
	public class WebMonitorSiteList : MarshalByRefObject
	{
		/// <summary>
		/// WebMonitor Site numeric Indexer
		/// </summary>
		public WebMonitorSite this[int Index]
		{
			get
			{
				return this.Sites[Index] as WebMonitorSite;
			}
			set
			{
				this.Sites[Index] = value;
			}
		}

		/// <summary>
		/// WebMonitor Site Indexer that returns a site by SiteId
		/// </summary>
		public WebMonitorSite this[string SiteId]
		{
			get
			{
				return this.GetSite(SiteId);
			}
			set
			{
				WebMonitorSite Site = this.GetSite(SiteId);
				Site = value;
			}
		}

		/// <summary>
		/// The count of sites
		/// </summary>
		public int Count
		{
			get { return this.Sites.Length; }
		}

		/// <summary>
		/// The XML file that contains the configuration.
		/// </summary>
		public string FileName;

		/// <summary>
		/// An error message string if negative returns.
		/// </summary>
		public string ErrorMsg = "";

		/// <summary>
		/// An error flag
		/// </summary>
		public bool Error = false;

		/// <summary>
		/// The core configuration object that provides the filenames
		/// and operational settings for requests.
		/// </summary>
		public WebMonitorConfig Config;

		/// <summary>
		/// An array of WebMonitor Sites that are managed by this list.
		/// Note: Not a collection, but an internal array member.
		/// </summary>
		public WebMonitorSite[] Sites;

		/// <summary>
		/// Flag that is used make Web Monitor stop checking.
		/// </summary>
		public bool lStopCheckSiteThread = true;

		/// <summary>
		/// This flag gets set if a new list is created. The IDE can check
		/// this and bring up Help in response.
		/// </summary>
		public bool IsNewList = false;

		//public WebMonitorMain oParentForm = null;
		private string cCodeBase = "";

		public WebMonitorSiteList()
		{
			this.FileName = Environment.CurrentDirectory + "\\WebMonitorSites.Xml";
			this.cCodeBase = this.GetType().Assembly.CodeBase;

			int lnLastSlash = this.cCodeBase.LastIndexOf("/");
			if ( lnLastSlash > 0 )
				this.cCodeBase = this.cCodeBase.Substring(0,lnLastSlash + 1);
			else 
			{
				lnLastSlash = this.cCodeBase.LastIndexOf("\\");
				this.cCodeBase = this.cCodeBase.Substring(0,lnLastSlash + 1);
			}

			//this.cConfigFilename = Environment.CurrentDirectory + "WebMonitorConfig.Xml";
		}

		/// <summary>
		/// Saves the current configuration to the WebMonitorSites.xml file (by default)
		/// </summary>
		/// <returns></returns>
		public bool Save() 
		{
			XmlTextWriter writer = null;
			try 
			{
				WebMonitorSerializer loWebMonitorSerializer = new WebMonitorSerializer();
				loWebMonitorSerializer.Sites = this.Sites;
				loWebMonitorSerializer.Config = this.Config;

				for (int x=0; x < this.Sites.Length; x++) 
				{
					if (this.Sites[x] != null) 
					{
						this.Sites[x].LastResponseText = null;
					}
				}
			
				XmlSerializer serializer = 
					new XmlSerializer(typeof(WebMonitorSerializer));

				// Create an XmlTextWriter using a FileStream.
				Stream fs = new FileStream(this.FileName, FileMode.Create);
				writer = 	new XmlTextWriter(fs, new UTF8Encoding());

				writer.Formatting = Formatting.Indented;
				writer.IndentChar = ' ';
				writer.Indentation = 4;
				
				// Serialize using the XmlTextWriter.
				serializer.Serialize(writer,loWebMonitorSerializer);
				writer.Close();
			}
			catch(Exception ex) 
			{
				this.ErrorMsg = ex.Message;
				this.Error = true;
			}
			finally
			{
				if (writer != null)
					writer.Close();
			}

			if (this.Error)			
				return false;

			return true;
		}

		/// <summary>
		/// Loads the Sites and oConfig members from disk. 
		/// </summary>
		/// <param name="llCreate">if true forces a file to be created if it doesn't exist</param>
		/// <returns></returns>
		public bool Load(bool llCreate) 
		{
		
			if (this.FileName == null) 
			{
				this.ErrorMsg = "No file specified...";
				return false;
			}

			FileInfo loFile = new FileInfo(this.FileName);
			if (!loFile.Exists) 
			{
				if (!llCreate) 
				{
					this.ErrorMsg = "File " + this.FileName + " doesn't exist.";
					return false;
				}

				/// *** Create new entries and save them to config file
				this.Config = new WebMonitorConfig();
				this.Sites = new WebMonitorSite[1];
				this.Sites[0] = new WebMonitorSite(this.Config);

				// Mark as new list
				this.IsNewList = true;
				
				try 
				{
					this.Save();
					return true;
				}
				catch(Exception) 
				{
					this.ErrorMsg = "Couldn't open or create " + this.FileName + ".";
					return false;
				}
			}

			XmlReader reader = null;
			XmlSerializer serializer = null;
			FileStream fs = null;

			try 
			{
				// Create an instance of the XmlSerializer specifying type and namespace.
				serializer = new XmlSerializer(typeof(WebMonitorSerializer));

				// A FileStream is needed to read the XML document.
				fs = new FileStream(this.FileName, FileMode.Open);
				reader = new XmlTextReader(fs);
				
				WebMonitorSerializer loWebMonitorSerializer = new WebMonitorSerializer();
			    
				// *** Use the Deserialize method to restore the object's state.
				loWebMonitorSerializer = (WebMonitorSerializer) serializer.Deserialize(reader);

				this.Config = loWebMonitorSerializer.Config;
				this.Sites = loWebMonitorSerializer.Sites;
				for (int x = 0; x < this.Sites.Length; x++) 
					if (this.Sites[x] != null)
						this.Sites[x].SetConfig(this.Config);
			}
			catch(Exception ex) 
			{
				this.ErrorMsg = ex.Message;
				return false;
			}
			finally
			{
				if (fs != null)
					fs.Close();

				if (reader != null)
					reader.Close();
			}

			return true;
		}

		/// <summary>
		/// Removes a site from the site list and repacks the SiteList Array.
		/// Note: this may affect any reference that is numeric.
		/// </summary>
		/// <param name="lcId"></param>
		/// <returns></returns>
		public bool Remove(string lcId) 
		{	
			if (this.Sites == null)
				return false;

			int lnSize = this.Sites.GetLength(0);
	
			WebMonitorSite loCurrent;

			int x;
			for ( x=0; x < lnSize; x++) 
			{
				loCurrent = this.Sites[x];
				if (loCurrent == null)
					continue;

				// *** if we match set to null and exit
				if (loCurrent.SiteId == lcId) 
				{
					this.Sites[x] = null;
					this.PackSites();
					return true;
				}
			}

			return false;
		}

		/// <summary>
		/// Packs the Sites array by removing any null references
		/// </summary>
		public void PackSites() 
		{
			int lnSize = 0;
			int lnOrigSize = this.Sites.GetLength(0);

			/// *** Find actual Size
			for(int x=0; x < lnOrigSize; x++) 
			{
				if (this.Sites[x]==null)
					continue;
				lnSize++;
			}

			WebMonitorSite[] TSites = new WebMonitorSite[lnSize];
			int lnNewSize = 0;
			for(int x=0; x < lnOrigSize ; x++) 
			{
				if (this.Sites[x]==null)
					continue;

				TSites[lnNewSize] = this.Sites[x];
				lnNewSize++;
			}

			this.Sites = TSites;
		}


		public void Add(WebMonitorSite loSite) 
		{
			if (this.Sites == null) 
			{
				this.Sites = new WebMonitorSite[1];
				this.Sites[0] = loSite;
				return;
			}

			int lnSize = this.Sites.GetLength(0);
			Array loTemp =  Array.CreateInstance(typeof(WebMonitorSite),this.Sites.GetLength(0) + 1);
			this.Sites.CopyTo(loTemp,0);

			// *** Now update the last item
			loTemp.SetValue(loSite,lnSize);

			this.Sites = (WebMonitorSite[]) loTemp;
		}

		/// <summary>
		/// Returns a Site by its SiteId from the array
		/// </summary>
		/// <param name="lcId"></param>
		/// <returns>WebMonitorSite</returns>
		public WebMonitorSite GetSite(string lcId) 
		{
			for (int x=0; x < this.Sites.GetLength(0); x++) 
			{
				if (this.Sites[x] == null)
					continue;

				if (this.Sites[x].SiteId == lcId)
					return this.Sites[x];
			}

			return null;
		}

		/// <summary>
		/// Returns a site by name.
		/// </summary>
		/// <param name="Name">The Site's Name</param>
		/// <returns></returns>
		public WebMonitorSite GetSiteByName(string Name) 
		{
			for (int x=0; x < this.Sites.GetLength(0); x++) 
			{
				if (this.Sites[x] == null)
					continue;

				if (this.Sites[x].SiteName.ToLower() == Name.ToLower() )
					return this.Sites[x];
			}

			return null;
		}

		/// <summary>
		/// Updates a Site in the SiteList Array
		/// </summary>
		/// <param name="loSite"></param>
		public void UpdateSite(WebMonitorSite loSite) 
		{

			for (int x=0; x < this.Sites.GetLength(0); x++) 
			{
				if (this.Sites[x] == null)
					continue;

				// *** If we have a match assign it
				if (this.Sites[x].SiteId == loSite.SiteId) 
				{
					this.Sites[x] = loSite; 
					return;
				}
			}
			
			// *** No match add it
			this.Add(loSite);
		}

		/// <summary>
		/// Runs through all sites and checks if it needs 
		/// to check the site.
		/// </summary>
		public void CheckSites()
		{
			// *** Loop through all the sites and check if they require firing
			for (int x=0; x < this.Sites.Length; x++ )
			{
				WebMonitorSite loSite = this.Sites[x];

				if (loSite == null)
					continue;

				if ( loSite.RequiresChecking() )
					loSite.CheckSiteAsync();  // starts new thread
			}
		}

		/// <summary>
		/// Fires up the background thread that loops through each of the sites
		/// </summary>
		public void CheckSitesAsync() 
		{
			// *** We're already running
			if (!this.lStopCheckSiteThread)
				return;

			// *** Start the background polling thread
			ThreadStart loEntryPoint = new ThreadStart(this.CheckSitesAsyncRun);
			Thread loThread = new Thread(loEntryPoint);
			loThread.Name = "SiteListMaster";
			loThread.Priority = ThreadPriority.BelowNormal;
			loThread.Start();
			
			this.lStopCheckSiteThread = false;
		}
	
		/// <summary>
		/// The background thread polling routine.
		/// This routine hooks up the event handlers to receive messages
		/// from the WebMonitorSite requests.
		/// Continuously calls the CheckSites method.
		/// </summary>
		void CheckSitesAsyncRun() 
		{
			// *** Hook up the Message Event Handler so we can receive events
			// *** Route all Site Events to this method (this.OnMessage)
			WebMonitorSite.OnMessageHandler Handler = 
				new WebMonitorSite.OnMessageHandler(this.OnMessage);

			// *** Initialize the event handler for each Site
			// *** NOTE: MUST initialize on this thread that is to receive event!
			for (int x = 0; x < this.Sites.Length; x++) 
				if (this.Sites[x] != null)
					this.Sites[x].OnMessage += Handler;

			int Loops = 0;

			// *** Flag that determines we need to stop checking
			this.lStopCheckSiteThread = false;

			while (true)
			{
				if (this.lStopCheckSiteThread)
					// *** Exit this thread routine
					return;

				this.CheckSites();
				Thread.Sleep(1000);			

				// *** Every 20 loops reduce the working set
				if (Loops % 20 == 0) 
				{
					wwUtils.SetWorkingSet(2000000,500000);
					Loops = 0;
				}

				Loops++;
			}

		}


		/// <summary>
		/// Attempts to shut down any pending threads of the
		/// existing running sites
		/// </summary>
		public void StopThreads() 
		{
			this.lStopCheckSiteThread = true;

			for (int x = 0; x < this.Sites.Length; x++ ) 
			{
				if (this.Sites[x] != null) 
				{
					this.Sites[x].StopThread();
				}
			}
		}


		public event delWebMonitorMessage WebMonitorMessage;
		public delegate void delWebMonitorMessage(WebMonitorSite Site, string Message, EventMessageId Id);

		/// <summary>
		/// Message event handler of the events that get fired from the Site. This intermediate
		/// handler captures the events and forwards them to the form's OnMessage handler with
		/// proper synchronization.
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="lcMessage"></param>
		/// <param name="lnId"></param>
		public void OnMessage(WebMonitorSite Site, string Message, EventMessageId Id) 
		{
			// *** Don't fire events if we're stopped
			if (this.lStopCheckSiteThread)
				return;

			if (this.WebMonitorMessage != null) 
			{
				// *** Fire the event directly - will fail eventually
				this.WebMonitorMessage(Site, Message,Id);
			}
		}

	}

	/// <summary>
	/// Class used for serialization
	/// </summary>
	public class WebMonitorSerializer 
	{
		/// <summary>
		/// A list of Web Monitor sites to monitor
		/// </summary>
		public WebMonitorSite[] Sites = null;

		/// <summary>
		/// The configuration object
		/// </summary>
		public WebMonitorConfig Config = null;
	}

	
}
	