using System;
using System.IO;

using Westwind.wwReflection.Tools;
using System.Runtime.InteropServices;
using System.Collections;

namespace Westwind.wwHelp
{
	/// <summary>
	/// The wwHelp class is an instance of a Help Builder Project. It's the 
	/// non-visual version of Help Builder that allows creation of projects, 
	/// compilation etc. This object is also available as a subobject of the 
	/// wwHelpForm object.
	/// 
	/// &lt;&lt;code lang=&quot;C#&quot;&gt;&gt;// *** Static Factory method
	/// wwHelp Help = wwHelp.CreateInstance();
	/// 
	/// Help.Open(&quot;d:\temp\test\test.hbp&quot;)
	/// 
	/// Help.LoadTopic(&quot;INDEX&quot;);
	/// Help.Topic.Topic = &quot;New Topic Title for this project&quot;
	/// Help.SaveTopic();
	/// 
	/// string Html = Help.Topic.RenderTopic();
	/// 
	/// Help.Close();
	/// 
	/// Help.Release();&lt;&lt;/code&gt;&gt;
	/// </summary>
	public class wwHelp : wwHelpComWrapper
	{
		/// <summary>
		/// Overridden constructor to allow passing a COM instance
		/// </summary>
		/// <param name="wwHelpInstance"></param>
		public wwHelp(object wwHelpInstance)
		{
			this.m_Inst = wwHelpInstance;
		}

		/// <summary>
		/// Static Factory method that creates an instance of this class.
		/// Instantiates the COM object and creates the wrapper for you.
		/// </summary>
		/// <returns></returns>
		public static wwHelp CreateInstance() 
		{
			try 
			{
				object HelpBuilder =  Activator.CreateInstance(Type.GetTypeFromProgID("wwHelp.wwHelp"));
				return new wwHelp(HelpBuilder);
			}
			catch
			{
			}
			return null;
		}


		/// <summary>
		/// Object containing all the detail for the currently active topic
		/// </summary>
		public wwHelpTopic Topic
		{
			get 
			{ 
					return new wwHelpTopic( wwUtils.GetPropertyCom(m_Inst,"oTopic") );
			}
		}

		/// <summary>
		/// Reference to the Help Form that was created with ShowForm
		/// </summary>
		public wwHelpForm HelpForm
		{
			get 
			{ 
				return new wwHelpForm (  wwUtils.GetPropertyCom(m_Inst,"oForm") );
			}
		}

		/// <summary>
		/// Opens a form and sets the HelpForm property to an instance of the form.
		/// </summary>
		/// <param name="Pk"></param>
		/// <param name="TopicType"></param>
		/// <returns></returns>
		public bool OpenForm(string Pk, TopicTypes TopicType) 
		{
			return (bool) wwUtils.CallMethodCom(this.m_Inst,"OpenForm",Pk,TopicType);
		}

		/// <summary>
		/// Opens a form and sets the HelpForm property to an instance of the form.
		/// </summary>
		public bool OpenForm(string Pk)
		{
			return this.OpenForm(Pk,TopicTypes.Pk);
		}

		/// <summary>
		/// Runs one of the Help Builder's Wizards. Not it opens a form and loads 
		/// up a new project if no form is open. Generally you should call this method 
		/// following a call to OpenForm() and whatever Open() commands on the form to
		/// load up a specific project.
		/// 
		/// Wizards available are:
		/// WizNewTopic
		/// WizBuildOutput
		/// WizNewProject
		/// WizImport
		/// </summary>
		/// <param name="WizardName">Name of the Wizard to run</param>
		public void RunWizard(string WizardName) 
		{
				wwUtils.CallMethodCom(this.m_Inst,"RunWizard",WizardName,false);
		}

		/// <summary>
		/// The full path to the currently open project file. This file can be empty in
		/// which case no project is open
		/// </summary>
		public string Filename
		{ 
			get { return (string) wwUtils.GetPropertyCom(m_Inst,"cFileName"); } 
			set { wwUtils.SetPropertyCom(m_Inst,"cFileName",value); }
		}

		/// <summary>
		/// Error Message that contains error information on various negative method
		/// results.
		/// </summary>
		public string ErrorMsg
		{ 
			get { return (string) wwUtils.GetPropertyCom(m_Inst,"cErrorMsg"); } 
			set { wwUtils.SetPropertyCom(m_Inst,"cErrorMsg",value); }
		}
		
		/// <summary>
		/// Error flag.
		/// </summary>
		public bool Error
		{ 
			get { return (bool) wwUtils.GetPropertyCom(m_Inst,"lError"); } 
			set { wwUtils.SetPropertyCom(m_Inst,"lError",value); }
		}

		/// <summary>
		/// Help Builders startup path or the path where the Help Builder
		/// application is currently running.
		/// </summary>
		public string AppStartPath 
		{
			get { return (string) wwUtils.GetPropertyCom(m_Inst,"cAppStartPath"); } 
			set { wwUtils.SetPropertyCom(m_Inst,"cAppStartPath",value); }
		}

		/// <summary>
		/// Name of the current Help Project
		/// </summary>
		public string ProjectName
		{ 
			get { return (string) wwUtils.GetPropertyCom(m_Inst,"cProjectName"); } 
			set { wwUtils.SetPropertyCom(m_Inst,"cProjectName",value); }
		}

		/// <summary>
		/// The language of the project in the format:
		/// 0x409 English (United States)
		/// 
		/// where the first word must be the hex country code. The rest is optional
		/// </summary>
		public string ProjectLanguage
		{
			get { return (string) wwUtils.GetPropertyCom(m_Inst,"cProjectLanguage"); }
			set { wwUtils.SetPropertyCom(m_Inst,"cProjectLanguage",value); }
		}

		/// <summary>
		/// Path where the generated HTML/CHM/DOC output is generated.
		/// </summary>
		public string OutputPath
		{
			get { return (string) wwUtils.GetPropertyCom(m_Inst,"cOutputPath"); }
			set { wwUtils.SetPropertyCom(m_Inst,"cOutputPath",value); }
		}

		
		/// <summary>
		/// The path where the project lives
		/// </summary>
		public string ProjectPath
		{
			get { return (string) wwUtils.GetPropertyCom(m_Inst,"cProjPath"); }
			set { wwUtils.SetPropertyCom(m_Inst,"cProjPath",value); }
		}
		
		/// <summary>
		/// The company name associcated with the current project 
		/// </summary>
		public string ProjectCompany
		{
			get { return (string) wwUtils.GetPropertyCom(m_Inst,"cProjCompany"); }
			set { wwUtils.SetPropertyCom(m_Inst,"cProjCompany",value); }
		}

		/// <summary>
		/// The Icon file associated with this project
		/// </summary>
		public string ProjectIconFile
		{
			get { return (string) wwUtils.GetPropertyCom(m_Inst,"cProjectIconFile"); }
			set { wwUtils.SetPropertyCom(m_Inst,"cProjectIconFile",value); }
		}

		/// <summary>
		/// Loads a topic based on a key. Topic is loaded into the Topic member.
		/// You can pass BLANK to get the topic to be blank.
		/// </summary>
		/// <param name="Pk">Can be PK, Signature, topic title depending on TopicType</param>
		/// <param name="TopicType">Type of lookup to perform</param>
		/// <returns></returns>
		public bool LoadTopic(string Pk, TopicTypes TopicType) 
		{
			int t = (int) TopicType;
			return (bool) wwUtils.CallMethodCom(m_Inst,"LoadTopic",Pk,t);
		}

		/// <summary>
		/// Loads a topic based on a key. Topic is loaded into the Topic member.
		/// You can pass BLANK to get the topic to be blank.
		/// </summary>
		/// <param name="Pk">PK of the topic to load</param>
		public bool LoadTopic(string Pk) 
		{
			return this.LoadTopic(Pk,TopicTypes.Pk);
		}

		/// <summary>
		/// Saves the currently active topic to disk
		/// </summary>
		/// <returns></returns>
		public bool SaveTopic( ) 
		{
			return (bool) wwUtils.CallMethodCom(m_Inst,"SaveTopic",false);
		}

		/// <summary>
		/// Creates a new topic and loads it into the Topic Member
		/// </summary>
		/// <returns></returns>
		public bool NewTopic() 
		{
			return (bool) wwUtils.CallMethodCom(m_Inst,"NewTopic");
		}

		/// <summary>
		/// Deletes the currently active topic from the table
		/// </summary>
		/// <param name="DeleteChildren"></param>
		public void DeleteTopic( bool DeleteChildren ) 
		{
			wwUtils.CallMethodCom(m_Inst,"DeleteTopic",DeleteChildren);
		}

		/// <summary>
		/// Checks to see if a topic exists in the current help file.
		/// </summary>
		/// <param name="Pk"></param>
		/// <returns></returns>
		public bool IsTopic(string Pk) 
		{
			return (bool) wwUtils.CallMethodCom(m_Inst,"IsTopic",Pk);
		}
		
		
		/// <summary>
		/// Saves the current project to a new filename.
		/// </summary>
		/// <param name="Filename"></param>
		/// <param name="IsBackup"></param>
		/// <returns></returns>
		public bool SaveTableAs( string Filename )
		{	
				return (bool) wwUtils.CallMethodCom(m_Inst,"SaveTableAs",false);
		
		}

		
		/// <summary>
		/// Opens an existing  project file.
		/// </summary>
		/// <param name="Filename"></param>
		/// <param name="Reopen"></param>
		/// <param name="Silent"></param>
		/// <returns></returns>
		public bool Open( string Filename, bool Reopen, bool Silent) 
		{
			return (bool) wwUtils.CallMethodCom(m_Inst,"Open",Filename,Reopen,Silent);			
		}

		/// <summary>
		/// Opens an existing  project file. This version is silent and fails
		/// if the project doesn't exist.
		/// </summary>
		public bool Open( string Filename) 
		{
			return this.Open(Filename,false,true);
		}

		/// <summary>
		/// Closes the currently open project file
		/// </summary>
		public void Close() 
		{
			wwUtils.CallMethodCom(m_Inst,"Close");
		}

		/// <summary>
		/// Sets the ErrorMesg property with this value. Pass null to clear
		/// </summary>
		/// <param name="Message"></param>
		public void SetError(string Message) 
		{
			wwUtils.CallMethodCom(m_Inst,"SetError",Message);
		}

		
		/// <summary>
		/// Renders the current topic into HTML
		/// </summary>
		/// <returns></returns>
		public string RenderTopic(string Pk) 
		{
			return (string) wwUtils.CallMethodCom(m_Inst,"RenderTopic",Pk);
		}
		/// <summary>
		/// Renders the current topic into HTML
		/// </summary>
		/// <returns></returns>
		public string RenderTopic() 
		{
			return this.RenderTopic("");
		}

		/// <summary>
		/// Low level function that renders Help Builder formatted text 
		/// into HTML. This function knows how to parse content containing
		/// double tags and RAWHTML attributes etc. Use this for taking
		/// content of fields like Body and rendering it into HTML.
		/// </summary>
		/// <param name="TopicText"></param>
		/// <param name="DontParseTopicLinks"></param>
		/// <returns></returns>
		public string FormatHtml(string TopicText) 
		{
			return (string) wwUtils.CallMethodCom(m_Inst,"FormatHtml",TopicText,false);
		}

		/// <summary>
		/// This routine creates a proper HREF link for the current view mode of the 
		/// control.
		/// </summary>
		/// <param name="Title"></param>
		/// <param name="Pk"></param>
		/// <param name="ExtraTags"></param>
		/// <param name="Anchor"></param>
		/// <returns></returns>
		public string GetHrefLink(string Title, string Pk, string ExtraTags, string Anchor) 
		{
				return (string) wwUtils.CallMethodCom(m_Inst,"GetHrefLink",Title,Pk,ExtraTags,Anchor);
		}

		/// <summary>
		/// This method finds the matching template for the current topic type.
		/// It returns the name of the template file.
		/// </summary>
		/// <param name="Template"></param>
		/// <returns></returns>
		public string SetTemplate(string TopicType) 
		{
			return (string) wwUtils.CallMethodCom(m_Inst,"SetTemplate",TopicType);
		}
			
			
		/// <summary>
		/// Returns an HTML formatted link list for Related topics. Basically takes
		/// the value of the SeeAlso field and converts it into HTML.
		/// </summary>
		/// <returns></returns>
		public string InsertSeeAlsoTopics() 
		{
			return (string) wwUtils.CallMethodCom(m_Inst,"InsertSeeAlsoTopics");
		}

		/// <summary>
		/// Returns the Inheritance Tree for the current topic
		/// </summary>
		/// <returns></returns>
		public string InsertInheritanceTree() 
		{
			return (string) wwUtils.CallMethodCom(m_Inst,"InsertInheritanceTree","");
		}
		
		/// <summary>
		/// Packs and Reindexes the Help File.
		/// </summary>
		public bool Reindex() 
		{	
			return (bool) wwUtils.CallMethodCom(m_Inst,"Reindex");
		}

		/// <summary>
		/// Creates a new project.
		/// </summary>
		/// <param name="ProjectFilename">Full path to a filename. Path or file doesn't need to exist Help Builder will create it.</param>
		/// <param name="Title">The title for the project.</param>
		/// <param name="Company">Optional company for the default copyright notice.</param>
		/// <returns></returns>
		public bool CreateProject(string ProjectFilename, string Title, string Company ) 
		{
			return (bool) wwUtils.CallMethodCom(m_Inst,"CreateProject",ProjectFilename, Title, Company);
		}


		/// <summary>
		/// Runs a query agains the Help Builder project and returns a DataRow 
		/// Collection.
		/// 
		/// &lt;&lt;code lang=&quot;C#&quot;&gt;&gt;
		/// wwHelp Hb = wwHelp.CreateInstance();
		/// Hb.Open(&quot;d:\temp\hbMsdn\hbMsdn&quot;);
		/// 
		/// System.Data.DataRowCollection Rows = 
		/// Hb.GetTopics(&quot;pk,topic,body&quot;,&quot;type='TOPIC'&quot;,&quot;PK&qu
		/// ot;);
		/// 
		/// Hb.Close();
		/// Hb.Release();
		/// &lt;&lt;/code&gt;&gt;
		/// <seealso>Class wwHelp</seealso>
		/// </summary>
		/// <param name="FieldList">
		/// list of fields to return (see docs)
		/// </param>
		/// <param name="WhereClause">
		/// Where clause SQL syntax
		/// </param>
		/// <param name="Order">
		/// Specify at least one sort order
		/// </param>
		/// <returns>DataRowCollection</returns>
		public System.Data.DataRowCollection GetTopics(string FieldList, string WhereClause, string Order )
		{	
			string Xml = (string) wwUtils.CallMethodCom(m_Inst,"GetTopics","XML",WhereClause,Order,FieldList,false);

			System.Data.DataSet ds = new System.Data.DataSet();
			ds.ReadXml( new StringReader(Xml) );

			if (ds.Tables.Count < 1 ) 
			{
				this.SetError("Invalid result set");
				return null;
			}

			 return ds.Tables[0].Rows;
		}

		/// <summary>
		/// Gets a dynamic property from the XML store.
		/// </summary>
		/// <param name="Property"></param>
		/// <returns></returns>
		public object GetProperty(string Property) 
		{
			return (bool) wwUtils.CallMethodCom(m_Inst,"GetProperty",Property,"");
		}

		/// <summary>
		/// Retrieves a dynamic value from the XML store. Note that VALUE should be 
		/// a simple value type only.
		/// </summary>
		/// <param name="Property"></param>
		/// <param name="Value"></param>
		public void SetProperty(string Property,object Value) 
		{
			wwUtils.CallMethodCom(m_Inst,"GetProperty",Property,Property,Value,"","");
		}

		/// <summary>
		/// Generates all the HTML output files for the current project
		/// </summary>
		/// <returns></returns>
		public bool GenerateHtml() 
		{
				return (bool) wwUtils.CallMethodCom(m_Inst,"GenerateHtml");
		}

		/// <summary>
		/// Generates the keyword index for the project
		/// </summary>
		/// <returns></returns>
		public bool GenerateIndex() 
		{
			return (bool) wwUtils.CallMethodCom(m_Inst,"GenerateIndex");
		}

		/// <summary>
		/// Generates the Content file for this and Content HTML pages for this 
		/// project.
		/// </summary>
		/// <param name="LimitSize">maximum number of topics allowed.</param>
		/// <param name="AllowServerSearch">Determines whether the Content page has links to an ASP.NET Search page option.</param>
		/// <returns></returns>
		/// LPARAMETER lnLimitSize, llAddDates, llSilent, lcIconFile, llAllowServerSearch

		public bool GenerateToc(int LimitSize,bool AllowServerSearch, string IconFile) 
		{
			return (bool) wwUtils.CallMethodCom(m_Inst,"GenerateToc",LimitSize,false,false,IconFile,AllowServerSearch);
		}

		/// <summary>
		/// Creates the actual CHM file. Note requires that HTML Workshop is installed.
		/// </summary>
		/// <returns></returns>
		public bool CompileProject()
		{
			return (bool) wwUtils.CallMethodCom(m_Inst,"CompileProject",false,"");
		}


	
#if false
		object addspecialrecord([In, Out, MarshalAs(UnmanagedType.Struct)] ref object lcType, [In, Out, MarshalAs(UnmanagedType.Struct)] ref object lcProjTitle, [In, Out, MarshalAs(UnmanagedType.Struct)] ref object lcCompany, [In, Out, MarshalAs(UnmanagedType.Struct)] ref object lcLanguage, [In, Out, MarshalAs(UnmanagedType.Struct)] ref object lcIconFile);
		object getimageidfromtopictype([In, Out, MarshalAs(UnmanagedType.Struct)] ref object lcType);
#endif
		
	}

}
